# from pathlib import Path

# import torch
# import pytest
# import lightning
# import util
# from torch.utils.data import DataLoader

# from torch_brain.data import Dataset
# from torch_brain.data.sampler import SequentialFixedWindowSampler
# from torch_brain.data.collate import collate
# from torch_brain.models import POYOPlus, POYOPlusTokenizer
# from torch_brain.taxonomy import decoder_registry
# from torch_brain.utils import train_wrapper, validation_wrapper

# import util

# DATA_ROOT = Path(util.get_data_paths()["processed_dir"]) / "processed"


# # this test only passes when only one gpu is visible
# def test_validation():
#     model = POYOPlus(
#         task_specs=decoder_registry,
#         backend_config="cpu",
#     )

#     tokenizer = POYOPlusTokenizer(
#         unit_tokenizer=model.unit_emb.tokenizer,
#         session_tokenizer=model.session_emb.tokenizer,
#         decoder_registry=decoder_registry,
#         latent_step=1.0 / 8,
#         num_latents_per_step=4,
#         batch_type=["stacked", "stacked", "stacked"],
#         eval=True,
#     )

#     ds = Dataset(
#         DATA_ROOT,
#         "valid",
#         [
#             {
#                 "selection": [
#                     {
#                         "dandiset": "mc_maze_small",
#                         "session": "jenkins_20090928_maze",
#                     }
#                 ],
#                 "config": {
#                     "multitask_readout": [
#                         {
#                             "decoder_id": "ARMVELOCITY2D",
#                             "weight": 2.0,
#                             "subtask_key": None,
#                             "metrics": [
#                                 {
#                                     "metric": "r2",
#                                     "task": "REACHING",
#                                 },
#                             ],
#                         }
#                     ],
#                 },
#             }
#         ],
#         transform=tokenizer,
#     )

#     model.unit_emb.initialize_vocab(ds.unit_ids)
#     model.session_emb.initialize_vocab(ds.session_ids)

#     sampler = SequentialFixedWindowSampler(
#         sampling_intervals=ds.get_sampling_intervals(),
#         window_length=1.0,
#     )
#     assert len(sampler) > 0

#     loader = DataLoader(ds, collate_fn=collate, batch_size=16, sampler=sampler)

#     device = "cpu"

#     model = model.to(device)

#     wrapper = train_wrapper.TrainWrapper(
#         model=model,
#         optimizer=None,
#         scheduler=None,
#     )
#     wrapper.to(device)

#     trainer = lightning.Trainer(accelerator=device)

#     validator = train_wrapper.CustomValidator(loader)
#     metrics_df = validator.on_validation_epoch_start(trainer, wrapper)

#     assert (
#         "val_mc_maze_small/jenkins_20090928_maze_armvelocity2d_r2"
#         == metrics_df.iloc[0].metric
#     )


# def avg_pool_naive(timestamps: torch.Tensor, values: torch.Tensor) -> torch.Tensor:
#     unique_timestamps, indices = torch.unique(timestamps, return_inverse=True)
#     averages = torch.zeros((len(unique_timestamps), *values.shape[1:]))

#     for i in range(len(unique_timestamps)):
#         group_values = values[indices == i]

#         if group_values.dtype == torch.long:
#             averages[i] = torch.mode(group_values, dim=0).values
#         else:
#             averages[i] = torch.mean(group_values, dim=0)
#     return averages


# def test_avg_pool():
#     # Test cases with different input shapes and data types
#     test_cases = [
#         # case: ground truth, float32. Should meanpool.
#         (
#             torch.tensor([0, 0, 1, 1, 2]),
#             torch.tensor([[1.5], [2.5], [3.9], [4.0], [5.0]]),
#         ),
#         # case: ground truth, int64. Should modepool.
#         (torch.tensor([0, 0, 1, 1, 2]), torch.tensor([[1], [2], [3], [4], [5]])),
#         # case: pred tensors with 2 classes. Should meanpool.
#         (
#             torch.tensor([0, 0, 1, 1, 2]),
#             torch.tensor([[1.0, 0.0], [0.0, 1.0], [1.0, 0.0], [0.0, 1.0], [1.0, 0.0]]),
#         ),
#     ]

#     for timestamps, values in test_cases:
#         # Fast implementation
#         fast_output = validation_wrapper.avg_pool(timestamps, values)
#         # Naive implementation
#         naive_output = avg_pool_naive(timestamps, values)
#         if values.dtype == torch.long:
#             fast_output = fast_output.long()
#             naive_output = naive_output.long()
#         # Compare the outputs
#         assert torch.allclose(
#             fast_output, naive_output, atol=1e-6
#         ), f"Fast and naive outputs do not match for inputs:\ntimestamps: {timestamps}\nvalues: {values}"


# def avg_pool_naive(timestamps: torch.Tensor, values: torch.Tensor) -> torch.Tensor:
#     unique_timestamps, indices = torch.unique(timestamps, return_inverse=True)
#     averages = torch.zeros((len(unique_timestamps), *values.shape[1:]))

#     for i in range(len(unique_timestamps)):
#         group_values = values[indices == i]

#         if group_values.dtype == torch.long:
#             averages[i] = torch.mode(group_values, dim=0).values
#         else:
#             averages[i] = torch.mean(group_values, dim=0)
#     return averages


# def test_avg_pool():
#     # Test cases with different input shapes and data types
#     test_cases = [
#         # case: ground truth, float32. Should meanpool.
#         (
#             torch.tensor([0, 0, 1, 1, 2]),
#             torch.tensor([[1.5], [2.5], [3.9], [4.0], [5.0]]),
#         ),
#         # case: ground truth, int64. Should modepool.
#         (torch.tensor([0, 0, 1, 1, 2]), torch.tensor([[1], [2], [3], [4], [5]])),
#         # case: pred tensors with 2 classes. Should meanpool.
#         (
#             torch.tensor([0, 0, 1, 1, 2]),
#             torch.tensor([[1.0, 0.0], [0.0, 1.0], [1.0, 0.0], [0.0, 1.0], [1.0, 0.0]]),
#         ),
#     ]

#     for timestamps, values in test_cases:
#         # Fast implementation
#         fast_output = validation_wrapper.avg_pool(timestamps, values)
#         # Naive implementation
#         naive_output = avg_pool_naive(timestamps, values)
#         if values.dtype == torch.long:
#             fast_output = fast_output.long()
#             naive_output = naive_output.long()
#         # Compare the outputs
#         assert torch.allclose(
#             fast_output, naive_output, atol=1e-6
#         ), f"Fast and naive outputs do not match for inputs:\ntimestamps: {timestamps}\nvalues: {values}"


# # def test_validation_willett():
# #     ds = Dataset(
# #         DATA_ROOT,
# #         "valid",
# #         OmegaConf.create([
# #             {
# #                 "selection": {
# #                     "dandiset": "willett_shenoy",
# #                     "session": "willett_shenoy_t5/t5.2019.05.08_single_letters",
# #                 },
# #                 "metrics": [{"output_key": "WRITING_CHARACTER"}],
# #             }
# #         ]),
# #     )
# #     batch_size = 16

# #     od = OrderedDict({x: 1 for x in ds.unit_names})
# #     vocab = torchtext.vocab.vocab(od, specials=["NA"])

# #     collate_fn = Collate(
# #         num_latents_per_step=16,  # This was tied in train_poyo_1.py
# #         step=1.0 / 8,
# #         sequence_length=1.0,
# #         unit_vocab=vocab,
# #         decoder_registry=decoder_registry,
# #         weight_registry=weight_registry,
# #     )

# #     model = PerceiverNM(
# #         unit_vocab=vocab,
# #         session_names=ds.session_names,
# #         num_latents=16,
# #         task_specs=decoder_registry
# #     )
# #     model = model.to('cuda')

# #     wrapper = train_wrapper.TrainWrapper(
# #         model=model,
# #         optimizer=None,
# #         scheduler=None,
# #     )
# #     wrapper.to('cuda')

# #     trainer = lightning.Trainer(accelerator="cuda")

# #     validator = train_wrapper.CustomValidator(ds, collate_fn)
# #     r2 = validator.on_validation_epoch_start(trainer, wrapper)
# #     assert "bce_writing_character" in r2.columns



import pytest
import numpy as np
import torch

from torch_brain.data import collate, pad, pad8, chain, track_mask, track_batch


def test_pad():
    # padding applied to np.ndarrays
    x = pad(np.array([1, 2, 3]))
    y = pad(np.array([4, 5]))

    batch = collate([x, y])
    assert torch.allclose(batch, torch.tensor([[1, 2, 3], [4, 5, 0]]))

    # padding applied to torch.Tensors
    x = pad(torch.tensor([[1], [2], [3]]))
    y = pad(torch.tensor([[4], [5]]))

    batch = collate([x, y])
    assert torch.allclose(batch, torch.tensor([[[1], [2], [3]], [[4], [5], [0]]]))

    # paddding applied to other objects (lists, maps, etc.)
    x = [pad({"a": np.array([1, 2, 3]), "b": np.array([11, 12, 13])}), np.array([0, 1])]
    y = [pad({"a": np.array([4, 5]), "b": np.array([14, 15])}), np.array([2, 3])]

    batch = collate([x, y])
    assert torch.allclose(batch[0]["a"], torch.tensor([[1, 2, 3], [4, 5, 0]]))
    assert torch.allclose(batch[0]["b"], torch.tensor([[11, 12, 13], [14, 15, 0]]))
    assert torch.allclose(batch[1], torch.tensor([[0, 1], [2, 3]]))


def test_pad8():
    # padding applied to np.ndarrays
    x = pad8(np.array([1, 2, 3]))
    y = pad8(np.array([4, 5]))

    batch = collate([x, y])
    assert torch.allclose(
        batch, torch.tensor([[1, 2, 3, 0, 0, 0, 0, 0], [4, 5, 0, 0, 0, 0, 0, 0]])
    )


def test_track_mask():
    # padding applied to np.ndarrays
    x_mask = track_mask(np.array([1, 2, 3]))
    y_mask = track_mask(np.array([4, 5]))

    batch = collate([x_mask, y_mask])
    assert batch.ndim == 2
    assert batch.dtype == torch.bool
    assert torch.allclose(batch, torch.BoolTensor([[1, 1, 1], [1, 1, 0]]))

    # padding applied to torch.Tensors
    x_mask = track_mask(torch.Tensor([[1], [2], [3]]))
    y_mask = track_mask(torch.Tensor([[4], [5]]))

    batch = collate([x_mask, y_mask])
    assert batch.ndim == 2
    assert batch.dtype == torch.bool
    assert torch.allclose(batch, torch.BoolTensor([[1, 1, 1], [1, 1, 0]]))

    # paddding applied to other objects (lists, maps, etc.)
    x = [
        pad({"a": np.array([1, 2, 3]), "b": np.array([11, 12, 13])}),
        track_mask(np.array([1, 2, 3])),
        np.array([0, 1]),
    ]
    y = [
        pad({"a": np.array([4, 5]), "b": np.array([14, 15])}),
        track_mask(np.array([4, 5])),
        np.array([2, 3]),
    ]

    batch = collate([x, y])
    assert batch[1].ndim == 2
    assert batch[1].dtype == torch.bool

    assert torch.allclose(batch[0]["a"], torch.tensor([[1, 2, 3], [4, 5, 0]]))
    assert torch.allclose(batch[0]["b"], torch.tensor([[11, 12, 13], [14, 15, 0]]))
    assert torch.allclose(batch[1], torch.BoolTensor([[1, 1, 1], [1, 1, 0]]))
    assert torch.allclose(batch[2], torch.tensor([[0, 1], [2, 3]]))


def test_chain():
    # chaining applied to np.ndarrays
    x = chain(np.array([1, 2, 3]))
    y = chain(np.array([4, 5]))

    batch = collate([x, y])
    assert torch.allclose(batch, torch.tensor([1, 2, 3, 4, 5]))

    # chaining applied to torch.Tensors
    x = chain(torch.tensor([[1], [2], [3]]))
    y = chain(torch.tensor([[4], [5]]))

    batch = collate([x, y])
    assert torch.allclose(batch, torch.tensor([[1], [2], [3], [4], [5]]))

    # chaining applied to other objects (lists, maps, etc.)
    x = [
        chain({"a": np.array([1, 2, 3]), "b": np.array([11, 12, 13])}),
        np.array([0, 1]),
    ]
    y = [chain({"a": np.array([4, 5]), "b": np.array([14, 15])}), np.array([2, 3])]
    batch = collate([x, y])

    assert torch.allclose(batch[0]["a"], torch.tensor([1, 2, 3, 4, 5]))
    assert torch.allclose(batch[0]["b"], torch.tensor([11, 12, 13, 14, 15]))
    assert torch.allclose(batch[1], torch.tensor([[0, 1], [2, 3]]))


def test_track_batch():
    # chaining applied to np.ndarrays
    x = track_batch(np.array([1, 2, 3]))
    y = track_batch(np.array([4, 5]))

    batch = collate([x, y])
    assert batch.ndim == 1
    assert batch.dtype == torch.int64
    assert torch.allclose(batch, torch.tensor([0, 0, 0, 1, 1]))

    # chaining applied to torch.Tensors
    x = track_batch(torch.tensor([[1], [2], [3]]))
    y = track_batch(torch.tensor([[4], [5]]))

    batch = collate([x, y])
    assert batch.ndim == 1
    assert batch.dtype == torch.int64
    assert torch.allclose(batch, torch.tensor([0, 0, 0, 1, 1]))

    # chaining applied to other objects (lists, maps, etc.)
    x = [
        chain({"a": np.array([1, 2, 3]), "b": np.array([11, 12, 13])}),
        track_batch(np.array([1, 2, 3])),
        np.array([0, 1]),
    ]
    y = [
        chain({"a": np.array([4, 5]), "b": np.array([14, 15])}),
        track_batch(np.array([4, 5])),
        np.array([2, 3]),
    ]
    batch = collate([x, y])

    assert torch.allclose(batch[0]["a"], torch.tensor([1, 2, 3, 4, 5]))
    assert torch.allclose(batch[0]["b"], torch.tensor([11, 12, 13, 14, 15]))
    assert torch.allclose(batch[1], torch.tensor([0, 0, 0, 1, 1]))
    assert torch.allclose(batch[2], torch.tensor([[0, 1], [2, 3]]))


def test_collate():
    # first sample
    a1 = np.array([1, 2, 3])
    b1 = np.array([[11, 12, 13], [14, 15, 16], [17, 18, 19]])
    c1 = np.array(
        [[[101, 102], [103, 104]], [[105, 106], [107, 108]], [[109, 110], [111, 112]]]
    )
    d1 = torch.tensor(1.0)
    e1 = torch.tensor([[1001, 1002], [1003, 1004]])
    data1 = dict(
        a=pad(a1),
        b=pad(b1),
        c=chain(c1),
        d=d1,
        e=e1,
        mask=track_mask(a1),
        batch=track_batch(c1),
    )

    # second sample
    a2 = np.array([4, 5])
    b2 = np.array([[20, 21, 22], [23, 24, 25]])
    c2 = np.array([[[113, 114], [115, 116]], [[117, 118], [119, 120]]])
    d2 = torch.tensor(2.0)
    e2 = torch.tensor([[1005, 1006], [1007, 1008]])
    data2 = dict(
        a=pad(a2),
        b=pad(b2),
        c=chain(c2),
        d=d2,
        e=e2,
        mask=track_mask(a2),
        batch=track_batch(c2),
    )

    # collate
    batch = collate([data1, data2])

    # check
    assert torch.allclose(batch["a"], torch.tensor([[1, 2, 3], [4, 5, 0]]))
    assert torch.allclose(
        batch["b"],
        torch.tensor(
            [
                [[11, 12, 13], [14, 15, 16], [17, 18, 19]],
                [[20, 21, 22], [23, 24, 25], [0, 0, 0]],
            ]
        ),
    )
    assert torch.allclose(
        batch["c"],
        torch.tensor(
            [
                [[101, 102], [103, 104]],
                [[105, 106], [107, 108]],
                [[109, 110], [111, 112]],
                [[113, 114], [115, 116]],
                [[117, 118], [119, 120]],
            ]
        ),
    )
    assert torch.allclose(batch["d"], torch.tensor([1.0, 2.0]))
    assert torch.allclose(
        batch["e"],
        torch.tensor([[[1001, 1002], [1003, 1004]], [[1005, 1006], [1007, 1008]]]),
    )
    assert batch["mask"].ndim == 2
    assert batch["mask"].dtype == torch.bool
    assert torch.allclose(batch["mask"], torch.BoolTensor([[1, 1, 1], [1, 1, 0]]))
    assert batch["batch"].ndim == 1
    assert batch["batch"].dtype == torch.int64
    assert torch.allclose(batch["batch"], torch.tensor([0, 0, 0, 1, 1]))


def test_chain_with_missing_keys():
    # chaining applied to np.ndarrays
    x = chain({"a": np.array([1, 2, 3])}, allow_missing_keys=True)
    y = chain({"a": np.array([4, 5]), "b": np.array([14, 15])}, allow_missing_keys=True)

    batch = collate([x, y])
    assert torch.allclose(batch["a"], torch.tensor([1, 2, 3, 4, 5]))
    assert torch.allclose(batch["b"], torch.tensor([14, 15]))

    # chaining should fail if keys are missing
    y = chain({"a": np.array([1, 2, 3])})
    x = chain({"a": np.array([4, 5]), "b": np.array([14, 15])})

    with pytest.raises(KeyError):
        collate([x, y])

    # chaining with allow_missing_keys=True should only work on dicts
    with pytest.raises(TypeError):
        x = chain(np.array([1, 2, 3]), allow_missing_keys=True)



import warnings
from collections.abc import Iterable
from typing import List, Union
from collections import OrderedDict

import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.parameter import UninitializedParameter


class InfiniteVocabEmbedding(nn.Module):
    """Embedding layer with a vocabulary that can be extended. Vocabulary is saved along
    with the model, and is reloaded when the state_dict is loaded. This is useful when
    the vocabulary is dynamically generated, e.g. from a dataset. For this reason this
    class also plays the role of the tokenizer.

    This layer is initially lazy, i.e. it does not have a weight matrix. The weight
    matrix is initialized when:

    - The vocabulary is initialized via :meth:`initialize_vocab()`.
    - or The model is loaded from a checkpoint that contains the vocabulary.

    If the vocabulary is initialized before :meth:`load_state_dict` is called,
    an error will be raised if the vocabulary in the checkpoint does not match the
    vocabulary in the model. The order of the words in the vocabulary does not matter,
    as long as the words are the same.

    If you would like to create a new variant of an existing :obj:`InfiniteVocabEmbedding`
    (that you loaded from a checkpoint), you can use:

    - :meth:`extend_vocab()` to add new words to the vocabulary. The embeddings for the new
    words will be initialized randomly.

    - :meth:`subset_vocab()` to select a subset of the vocabulary. The embeddings for the
    selected words will be copied from the original embeddings, and the ids for the
    selected words will change and :meth:`tokenizer` will be updated accordingly.

    This module also plays the role of the tokenizer, which is accessible via
    :meth:`tokenizer`, and is a Callable.

    .. warning:: If you are only interested in loading a subset of words from a checkpoint, do not call :meth:`initialize_vocab()`, first load the checkpoint then use :meth:`subset_vocab`.

    Args:
        embedding_dim (int): Embedding dimension.
        init_scale (float): The standard deviation of the normal distribution used to
            initialize the embedding matrix. Default is 0.02.
    """

    def __init__(self, embedding_dim, init_scale=0.02):
        super().__init__()

        self.embedding_dim = embedding_dim
        self.init_scale = init_scale
        self.padding_idx = 0

        self.weight = UninitializedParameter()
        self.vocab = None

        # Unfortunately, this hook is private, though there has been a PR to make it
        # public: https://github.com/pytorch/pytorch/issues/75287
        self._register_load_state_dict_pre_hook(
            self._hook_vocab_on_load_state_dict, with_module=False
        )

    def initialize_vocab(self, vocab: List[str]):
        r"""Initialize the vocabulary with a list of words. This method should be called
        only once, and before the model is trained. If you would like to add new words
        to the vocabulary, use :obj:`extend_vocab()` instead.

        .. note:: A special word "NA" will always be in the vocabulary, and is assigned the index 0. 0 is used for padding.

        Args:
            vocab (List[str]): A list of words to initialize the vocabulary.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)

            >>> embedding.vocab
            OrderedDict([('NA', 0), ('apple', 1), ('banana', 2), ('cherry', 3)])

            >>> embedding.weight.shape
            torch.Size([4, 64])
        """
        assert (
            self.vocab is None
        ), f"Vocabulary already initialized, and has {len(self.vocab)} words. "
        "If you want to add new words to the vocabulary, use extend_vocab() instead."

        # Create a mapping from words to indices
        if isinstance(vocab, str):
            raise ValueError("vocab cannot be a single string")
        elif isinstance(vocab, Iterable):
            # check that all words are unique
            if len(vocab) != len(set(vocab)):
                raise ValueError("Vocabulary contains duplicate words")
            self.vocab = OrderedDict(zip(vocab, range(1, len(vocab) + 1)))
            assert "NA" not in self.vocab, "NA is a reserved word"
            self.vocab["NA"] = 0
            self.vocab.move_to_end("NA", last=False)
        else:
            raise ValueError("vocab must be a list or dict")

        self.initialize_parameters(len(self.vocab))

    def extend_vocab(self, vocab: List[str], exist_ok=False):
        r"""Extend the vocabulary with a list of words. If a word already exists in the
        vocabulary, an error will be raised. The embeddings for the new words will be
        initialized randomly, and new ids will be assigned to the new words.

        Args:
            vocab (List[str]): A list of words to add to the vocabulary.
            exist_ok (bool): If True, the method will not raise an error if the new words
                already exist in the vocabulary and will skip them. Default is False.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)
            >>> embedding
            InfiniteVocabEmbedding(embedding_dim=64, num_embeddings=4)

            >>> new_words = ["date", "elderberry", "fig"]
            >>> embedding.extend_vocab(new_words)
            InfiniteVocabEmbedding(embedding_dim=64, num_embeddings=7)

            >>> embedding.vocab
            OrderedDict([('NA', 0), ('apple', 1), ('banana', 2), ('cherry', 3), ('date', 4), ('elderberry', 5), ('fig', 6)])

            >>> embedding.weight.shape
            torch.Size([7, 64])
        """
        if self.is_lazy():
            raise ValueError("No vocabulary was initialized. Use initialize_vocab()")

        # find intersection and difference between key sets
        new_words, existing_words = [], []
        for word in vocab:
            if word not in self.vocab:
                new_words.append(word)
            else:
                existing_words.append(word)

        if not exist_ok and len(existing_words) > 0:
            raise ValueError(
                f"Vocabulary already contains {len(existing_words)} out of {len(vocab)}"
                f" words that are being added. You can skip this error by setting "
                f"exist_ok=True, but be aware that the embeddings for these existing "
                f"words won't be re-initialized."
            )

        # update tokenizer
        self.vocab.update(
            OrderedDict(
                zip(new_words, range(len(self.vocab), len(self.vocab) + len(new_words)))
            )
        )

        # check that the dictionary update was done correctly
        assert len(self.vocab) == len(self.weight) + len(new_words), (
            f"Expected vocabulary length to be {len(self.weight) + len(new_words)}, "
            f"but got {len(self.vocab)}"
        )

        # check that the largest value in the vocab matches its size
        assert max(self.vocab.values()) == len(self.vocab) - 1, (
            f"Expected maximum vocabulary index to be {len(self.vocab) - 1}, "
            f"but got {max(self.vocab.values())}"
        )

        # make a copy of existing embeddings
        embeddings_for_existing_words = self.weight.clone().detach()

        # reinitalize weight matrix after extending it
        self.weight = UninitializedParameter()
        self.initialize_parameters(len(self.vocab))

        # copy existing embeddings into new weight matrix
        self.weight.data[: len(embeddings_for_existing_words)] = (
            embeddings_for_existing_words
        )
        return self

    def subset_vocab(self, vocab: List[str], inplace=True):
        r"""Select a subset of the vocabulary. The embeddings for the selected words
        will be copied from the original embeddings, and the ids for the selected words
        will be updated accordingly.

        An error will be raised if one of the words does not exist in the vocabulary.

        Args:
            vocab (List[str]): A list of words to select from the vocabulary.
            inplace (bool): If True, the method will modify the vocabulary and the weight
                matrix in place. If False, a new InfiniteVocabEmbedding will be returned
                with the selected words. Default is True.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)
            >>> embedding
            InfiniteVocabEmbedding(embedding_dim=64, num_embeddings=4)

            >>> selected_words = ["banana", "cherry"]
            >>> embedding.subset_vocab(selected_words)
            InfiniteVocabEmbedding(embedding_dim=64, num_embeddings=3)

            >>> embedding.vocab
            OrderedDict([('NA', 0), ('banana', 1), ('cherry', 2)])

            >>> embedding.weight.shape
            torch.Size([3, 64])
        """
        if self.is_lazy():
            raise ValueError("No vocabulary was initialized. Use initialize_vocab()")

        assert len(vocab) > 0, "Vocabulary must contain at least one word."

        # check that all words are unique
        if len(vocab) != len(set(vocab)):
            raise ValueError("Vocabulary contains duplicate words")

        # find intersection and difference between key sets
        word_indices = [0]  # NA will be added
        for word in vocab:
            if word not in self.vocab:
                raise ValueError(f"Vocabulary does not contain word {word}")
            else:
                word_indices.append(self.tokenizer(word))

        # make a copy of selected embeddings
        word_indices = torch.tensor(word_indices, dtype=torch.long)
        with torch.no_grad():
            embeddings_for_selected_words = self.weight[word_indices].clone().detach()

        if inplace:
            self.vocab = None
            self.weight = UninitializedParameter()

            self.initialize_vocab(vocab)

            self.weight.data = embeddings_for_selected_words
            return self
        else:
            new_embedding = self.__class__(self.embedding_dim, self.init_scale)
            new_embedding.initialize_vocab(vocab)
            new_embedding.weight.data = embeddings_for_selected_words
            return new_embedding

    def tokenizer(self, words: Union[str, List[str]]):
        r"""Convert a word or a list of words to their token indices.

        Args:
            words (Union[str, List[str]]): A word or a list of words.

        Returns:
            Union[int, List[int]]: A token index or a list of token indices.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)

            >>> embedding.tokenizer("banana")
            2

            >>> embedding.tokenizer(["apple", "cherry", "apple"])
            [1, 3, 1]
        """
        if self.is_lazy():
            raise ValueError("No vocabulary was initialized. Use initialize_vocab()")

        if isinstance(words, str):
            return self.vocab[words]
        return [self.vocab[w] for w in words]

    def detokenizer(self, index: int):
        r"""Convert a token index to a word.

        Args:
            index (int): A token index.

        Returns:
            str: A word.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)

            >>> embedding.detokenizer(2)
            'banana'
        """
        if self.is_lazy():
            raise ValueError("No vocabulary was initialized. Use initialize_vocab()")

        return list(self.vocab.keys())[index]

    def is_lazy(self):
        r"""Returns True if the module is not initialized.

        Example ::

            >>> from torch_brain.nn import InfiniteVocabEmbedding

            >>> embedding = InfiniteVocabEmbedding(64)

            >>> embedding.is_lazy()
            True

            >>> vocab = ["apple", "banana", "cherry"]
            >>> embedding.initialize_vocab(vocab)

            >>> embedding.is_lazy()
            False
        """
        return isinstance(self.weight, UninitializedParameter)

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module, but will not reset the
        vocabulary."""
        if not self.is_lazy():
            torch.nn.init.normal_(self.weight, mean=0, std=self.init_scale)
            if self.padding_idx is not None:
                with torch.no_grad():
                    self.weight[self.padding_idx].fill_(0)

    def initialize_parameters(self, num_embeddings) -> None:  # type: ignore[override]
        if self.is_lazy():
            with torch.no_grad():
                self.weight.materialize((num_embeddings, self.embedding_dim))
                self.reset_parameters()

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        if self.is_lazy():
            destination[prefix + "weight"] = self.weight
            destination[prefix + "vocab"] = self.vocab
        else:
            super()._save_to_state_dict(destination, prefix, keep_vars)
            destination[prefix + "vocab"] = self.vocab

    def _hook_vocab_on_load_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        if not self.is_lazy():
            # pop the vocabulary from the state_dict
            # popping is done because by default pytorch does not know how to load
            # vocab which is an OrderedDict, so we cannot keep it inside of the state_dict.
            # however, if we pop it, it will also be removed in the original state_dict
            # which can have unintended side effects, noteably this will alter the actual
            # state_dict in the checkpoint or the other module being used.
            # TODO: find a way to keep the vocab in the state_dict
            incoming_vocab = state_dict.pop(prefix + "vocab")

            # incoming_vocab and self.vocab might have the same keys but in a different order
            # reorder incoming_vocab to match self.vocab, and get the remapping indices
            remap_indices = []
            for word, index in self.vocab.items():
                if not word in incoming_vocab:
                    raise ValueError(
                        f"Vocabulary mismatch: word {word} is missing. If "
                        "you would like to add new words, or a new "
                        "vocabulary, do not initialize the vocab, load the"
                        " checkpoint, and then call extend_vocab() to add"
                        "new words, and/or subset_vocab() to remove words."
                    )
                remap_indices.append(incoming_vocab.pop(word))
            if len(incoming_vocab) > 0:
                raise ValueError(
                    f"Vocabulary mismatch: {len(incoming_vocab)} words are"
                    "remaining and cannot be loaded. If you would like to "
                    "load a subset of the vocabulary, do not initialize "
                    "the vocab, load the checkpoint, and then call "
                    "subset_vocab()"
                )
            remap_indices = torch.tensor(remap_indices, dtype=torch.long)

            state_dict[prefix + "weight"] = state_dict[prefix + "weight"][remap_indices]

        else:
            if not isinstance(state_dict[prefix + "weight"], UninitializedParameter):
                # The module is not initialized, but the one being loaded is
                with torch.no_grad():
                    self.weight.materialize(state_dict[prefix + "weight"].shape)
                self.vocab = state_dict.pop(prefix + "vocab")
            else:
                # Both the module and the one being loaded are not initialized
                assert state_dict.pop(prefix + "vocab") is None

    def forward(self, input):
        if self.is_lazy():
            raise ValueError("No vocabulary was initialized. Use initialize_vocab()")
        return F.embedding(input, self.weight, self.padding_idx)

    def extra_repr(self) -> str:
        return "embedding_dim={}, num_embeddings={}".format(
            self.embedding_dim, len(self.vocab) if self.vocab is not None else 0
        )



from typing import Optional

import torch
import torch.nn.functional as F
import torch.nn as nn
from einops import rearrange, repeat

try:
    import xformers.ops as xops
except ImportError:
    xops = None


from torch_brain.nn.rotary_embedding import apply_rotary_pos_emb


class RotaryCrossAttention(nn.Module):
    """Cross-attention layer with rotary positional embeddings.

    This layer performs cross-attention between a query sequence and a context sequence,
    with rotary positional embeddings applied to the queries and keys (and optionally values).
    It first normalizes the inputs, projects them to query/key/value space, applies rotary
    embeddings and attention, then projects back to the original dimension.

    The layer provides two forward methods:

    - forward(): This is the default, and is used for sequences in a batch that are of
      the same length, or are padded to the same length. When padding is used, attention
      masks need to be provided.
    - forward_varlen(): Uses sequence lengths instead of masks for sequences that are chained
      together in a single batch dimension. This can be more memory efficient since it avoids
      padding, but requires the sequences to be concatenated rather than stacked.

    Args:
        dim (int): Dimension of input query embeddings
        context_dim (Optional[int]): Dimension of input context embeddings. If None, uses same as dim
        heads (int): Number of attention heads
        dim_head (int): Dimension of each attention head
        dropout (float): Dropout probability
        rotate_value (bool): Whether to apply rotary embeddings to values as well as queries/keys
    """

    def __init__(
        self,
        *,
        dim: int,
        context_dim: Optional[int] = None,
        heads: int = 8,
        dim_head: int = 64,
        dropout: float = 0.0,
        rotate_value: bool = False,
    ):
        super().__init__()

        inner_dim = dim_head * heads
        context_dim = context_dim or dim
        self.heads = heads
        self.dropout = dropout
        self.rotate_value = rotate_value

        self.norm = nn.LayerNorm(dim)
        self.norm_context = nn.LayerNorm(context_dim)

        self.to_q = nn.Linear(dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)
        self.to_out = nn.Linear(inner_dim, dim)

    def forward(
        self,
        x_query,
        x_context,
        query_pos_emb,
        context_pos_emb,
        context_mask=None,
    ):
        """Forward pass for regular or padded sequences.

        Shape:
            - x_query: (B, N_q, D_q)
            - x_context: (B, N_c, D_c)
            - query_pos_emb: (B, N_q, D_h)
            - context_pos_emb: (B, N_c, D_h)
            - context_mask: Optional[Tensor[B, N_c]]
            - Output: (B, N_q, D)

            where B is batch size, N_q is query sequence length, N_c is context sequence
            length, D_q is input dimension, D_c is context dimension, H is number of heads,
            and D_h is head dimension.
        """
        # normalize
        x_query = self.norm(x_query)
        x_context = self.norm_context(x_context)

        # project to q, k, v
        q = self.to_q(x_query)
        k, v = self.to_kv(x_context).chunk(2, dim=-1)

        # select attention kernel
        if xops is not None and x_query.device.type == "cuda":
            # if xformers is available, use it for attention.
            # xformers supports attention masks when using the memory efficient attention
            # kernel, but pytorch does not.
            rotary_attn_func = rotary_attn_xformers_func
        else:
            # otherwise use pytorch's default attention which will determine the best
            # attention kernel (math, mem_efficient or flash) based on the hardware and
            # other factors.
            rotary_attn_func = rotary_attn_pytorch_func

        # apply attention
        out = rotary_attn_func(
            query=q,
            key=k,
            value=v,
            q_pos_emb=query_pos_emb,
            kv_pos_emb=context_pos_emb,
            num_heads=self.heads,
            dropout_p=self.dropout if self.training else 0,
            rotate_value=self.rotate_value,
            attn_mask=context_mask,
        )

        # project back to dim
        out = self.to_out(out)
        return out

    def forward_varlen(
        self,
        x_query,
        x_context,
        query_pos_emb,
        context_pos_emb,
        query_seqlen,
        context_seqlen,
    ):
        """Forward pass for variable length sequences.

        Similar to forward() but handles variable length sequences that have been chained
        together in the batch dimension rather than being stacked and padded. This approach
        can be more memory efficient since it avoids padding, but requires the sequences
        to be concatenated rather than stacked.

        Shape:
            - x_query: (N_q_total, D)
            - x_context: (N_c_total, D_c)
            - query_pos_emb: (N_q_total, D_h)
            - context_pos_emb: (N_c_total, D_h)
            - query_seqlen: (B,)
            - context_seqlen: (B,)
            - Output: (N_q_total, D)

            where N_q_total and N_c_total are the total sequence lengths across the batch,
            B is batch size, D is input dimension, D_c is context dimension, H is number of
            heads, and D_h is head dimension.
        """
        # normalize
        x_query = self.norm(x_query)
        x_context = self.norm_context(x_context)

        # project to q, k, v
        q = self.to_q(x_query)
        k, v = self.to_kv(x_context).chunk(2, dim=-1)

        # select attention kernel
        if xops is not None and x_query.device.type == "cuda":
            rotary_attn_func = rotary_attn_xformers_varlen_func
        else:
            if x_query.device.type == "cuda":
                raise RuntimeError(
                    "No varlen attention kernel available, please install xformers."
                )
            else:
                # forward_varlen is not implemented for CPU, forward should be used instead
                raise NotImplementedError(
                    "No varlen attention kernel available for CPU."
                )

        # apply attention
        out = rotary_attn_func(
            query=q,
            key=k,
            value=v,
            q_pos_emb=query_pos_emb,
            kv_pos_emb=context_pos_emb,
            num_heads=self.heads,
            dropout_p=self.dropout if self.training else 0,
            rotate_value=self.rotate_value,
            q_seqlen=query_seqlen,
            kv_seqlen=context_seqlen,
        )

        # project back to dim
        out = self.to_out(out)
        return out


class RotarySelfAttention(nn.Module):
    """Self-attention layer with rotary positional embeddings.

    This layer performs self-attention within a sequence, with rotary positional embeddings
    applied to the queries and keys (and optionally values). It first normalizes the input,
    projects it to query/key/value space, applies rotary embeddings and attention, then
    projects back to the original dimension.

    The layer provides two forward methods:
    - forward(): This is the default, and is used for sequences in a batch that are of
      the same length, or are padded to the same length. When padding is used, attention
      masks need to be provided.
    - forward_varlen(): Uses sequence lengths instead of masks for sequences that are chained
      together in a single batch dimension. This can be more memory efficient since it avoids
      padding, but requires the sequences to be concatenated rather than stacked.

    Args:
        dim (int): Dimension of input embeddings
        heads (int): Number of attention heads
        dim_head (int): Dimension of each attention head
        dropout (float): Dropout probability
        rotate_value (bool): Whether to apply rotary embeddings to values as well as queries/keys
    """

    def __init__(
        self,
        *,
        dim: int,
        heads: int = 8,
        dim_head: int = 64,
        dropout: float = 0.0,
        rotate_value: bool = False,
    ):
        super().__init__()

        inner_dim = dim_head * heads
        self.heads = heads
        self.dropout = dropout
        self.rotate_value = rotate_value

        self.norm = nn.LayerNorm(dim)

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim)

    def forward(
        self,
        x,
        rotary_time_emb,
        x_mask=None,
    ):
        """Forward pass for fixed-length sequences.

        Shape:
            - x: (B, N, D)
            - rotary_time_emb: (B, N, D_h)
            - x_mask: (B, N, N)
            - Output: (B, N, D)

            where B is batch size, N is sequence length, D is input dimension,
            and D_h is head dimension.
        """
        # normalize
        x = self.norm(x)

        # project to q, k, v
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)

        # select attention kernel
        if xops is not None and x.device.type == "cuda":
            rotary_attn_func = rotary_attn_xformers_func
        else:
            rotary_attn_func = rotary_attn_pytorch_func

        # apply attention
        out = rotary_attn_func(
            query=q,
            key=k,
            value=v,
            q_pos_emb=rotary_time_emb,
            kv_pos_emb=rotary_time_emb,
            num_heads=self.heads,
            dropout_p=self.dropout if self.training else 0,
            rotate_value=self.rotate_value,
            attn_mask=x_mask,
        )

        # project back to dim
        out = self.to_out(out)
        return out

    def forward_varlen(
        self,
        x,
        rotary_time_emb,
        x_seqlen,
    ):
        """Forward pass for variable-length sequences.

        Shape:
            - x: (N_total, D)
            - rotary_time_emb: (N_total, D_h)
            - x_seqlen: (B,)
            - Output: (N_total, D)

            where N_total is the total sequence length across the batch,
            B is batch size, D is input dimension, and D_h is head dimension.
        """
        # normalize
        x = self.norm(x)

        # project to q, k, v
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)

        # select attention kernel
        if xops is not None and x.device.type == "cuda":
            rotary_attn_func = rotary_attn_xformers_varlen_func
        else:
            if x.device.type == "cuda":
                raise RuntimeError(
                    "No varlen attention kernel available, please install xformers."
                )
            else:
                # forward_varlen is not implemented for CPU, forward should be used instead
                raise NotImplementedError(
                    "No varlen attention kernel available for CPU."
                )

        # apply attention
        out = rotary_attn_func(
            query=q,
            key=k,
            value=v,
            q_pos_emb=rotary_time_emb,
            kv_pos_emb=rotary_time_emb,
            num_heads=self.heads,
            dropout_p=self.dropout if self.training else 0,
            rotate_value=self.rotate_value,
            q_seqlen=x_seqlen,
            kv_seqlen=None,  # self-attention has the same seqlen for q, k, v
        )

        # project back to dim
        out = self.to_out(out)
        return out


def rotary_attn_pytorch_func(
    *,
    query,
    key,
    value,
    q_pos_emb,
    kv_pos_emb,
    attn_mask=None,
    num_heads: int,
    dropout_p: float,
    rotate_value: bool,
):
    # uses the default scaled dot product attention from pytorch
    # https://pytorch.org/docs/stable/generated/torch.nn.functional.scaled_dot_product_attention.html
    # this implements basic versions of memory efficient attention and flash attention
    # but more advanced versions are available in xformers and flash_attn (varlen)
    # which allow us to perform complex masking operations
    r"""Wraps the default attention implementation with rotary embedding application.

    Args:
        query: The query tensor, with shape (b, n_q, (h d))
        key: The key tensor, with shape (b, n_kv, (h d))
        value: The value tensor, with shape (b, n_kv, (h d))
        q_pos_emb: The query rotary position embedding, with shape (b, n_q, d)
        kv_pos_emb: The key rotary position embedding, with shape (b, n_kv, d)
        num_heads: The number of attention heads
        dropout_p: The dropout probability
        rotate_value: Whether to rotate the value in addition to the query and key
        attn_mask: The attention mask, with shape (b, n_kv)

    Returns:
        The output tensor, with shape (b, n_q, (h d))
    """

    # default attention expects shape b h n d
    query = rearrange(query, "b n (h d) -> b h n d", h=num_heads)
    key = rearrange(key, "b n (h d) -> b h n d", h=num_heads)
    value = rearrange(value, "b n (h d) -> b h n d", h=num_heads)

    # apply rotary embeddings
    query = apply_rotary_pos_emb(q_pos_emb, query, head_dim=1)
    key = apply_rotary_pos_emb(kv_pos_emb, key, head_dim=1)
    if rotate_value:
        value = apply_rotary_pos_emb(kv_pos_emb, value, head_dim=1)

    # attention mask
    if attn_mask is not None:
        attn_mask = rearrange(attn_mask, "b n -> b () () n")

    # perform attention, by default will use the optimal attention implementation
    out = F.scaled_dot_product_attention(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
    )

    if rotate_value:
        out = apply_rotary_pos_emb(-q_pos_emb, out, head_dim=1)

    # return (b, n, (h d), )
    out = rearrange(out, "b h n d -> b n (h d)")
    return out


def rotary_attn_xformers_func(
    *,
    query,
    key,
    value,
    q_pos_emb,
    kv_pos_emb,
    attn_mask=None,
    num_heads: int,
    dropout_p: float,
    rotate_value: bool,
):
    r"""Wraps the memory efficient attention implementation with rotary embedding
    application.

    Args:
        query: The query tensor, with shape (b n (h d))
        key: The key tensor, with shape (b n (h d))
        value: The value tensor, with shape (b n (h d))
        q_pos_emb: The query rotary position embedding, with shape (b n d)
        kv_pos_emb: The key rotary position embedding, with shape (b n d)
        attn_mask: The attention mask, with shape (b, n_kv). A value of True indicates
            that the element should take part in attention.
        num_heads: The number of attention heads
        dropout_p: The dropout probability
        rotate_value: Whether to rotate the value in addition to the query and key

    Returns:
        The output tensor, with shape (b n (h d))
    """
    # xformers attention expects shape (1, n, h, d)
    query = rearrange(query, "b n (h d) -> b n h d", h=num_heads)
    key = rearrange(key, "b n (h d) -> b n h d", h=num_heads)
    value = rearrange(value, "b n (h d) -> b n h d", h=num_heads)

    query = apply_rotary_pos_emb(q_pos_emb, query, head_dim=2)
    key = apply_rotary_pos_emb(kv_pos_emb, key, head_dim=2)

    if rotate_value:
        value = apply_rotary_pos_emb(kv_pos_emb, value, head_dim=2)

    # WARNING: this is very slow, avoid using attn_mask if possible, refer to xformers
    # documentation
    attn_mask = (
        repeat(attn_mask, "b m -> b h n m", h=num_heads, n=query.size(1))
        if attn_mask is not None
        else None
    )
    attn_bias = (
        attn_mask.to(query.dtype).masked_fill(attn_mask.logical_not(), float("-inf"))
        if attn_mask is not None
        else None
    )

    out = xops.memory_efficient_attention(
        query,
        key,
        value,
        attn_bias=attn_bias,
        p=dropout_p,
    )

    if rotate_value:
        out = apply_rotary_pos_emb(-q_pos_emb, out, head_dim=2)

    out = rearrange(out, "b n h d -> b n (h d)")
    return out


def rotary_attn_xformers_varlen_func(
    *,
    query,
    key,
    value,
    q_pos_emb,
    kv_pos_emb,
    q_seqlen,
    kv_seqlen,
    num_heads: int,
    dropout_p: float,
    rotate_value: bool,
):
    r"""Wraps the memory efficient attention implementation with rotary embedding
    application.

    Args:
        query: The query tensor, with shape (n, (h d))
        key: The key tensor, with shape (n, (h d))
        value: The value tensor, with shape (n, (h d))
        query_pos_emb: The query rotary position embedding, with shape (n, d)
        key_pos_emb: The key rotary position embedding, with shape (n, d)
        num_heads: The number of attention heads
        dropout_p: The dropout probability
        rotate_value: Whether to rotate the value in addition to the query and key
        q_seqlen: The sequence length of the query tensor
        kv_seqlen: The sequence length of the key and value tensors

    Returns:
        The output tensor, with shape (n, (h d))
    """
    # xformers attention expects shape (1, n, h, d)
    query = rearrange(query, "n (h d) -> () n h d", h=num_heads)
    key = rearrange(key, "n (h d) -> () n h d", h=num_heads)
    value = rearrange(value, "n (h d) -> () n h d", h=num_heads)

    # TODO check rotation works
    query = apply_rotary_pos_emb(q_pos_emb.unsqueeze(0), query)
    key = apply_rotary_pos_emb(kv_pos_emb.unsqueeze(0), key)

    if rotate_value:
        value = apply_rotary_pos_emb(kv_pos_emb.unsqueeze(0), value)

    if isinstance(q_seqlen, torch.Tensor):
        q_seqlen = q_seqlen.tolist()
    if isinstance(kv_seqlen, torch.Tensor):
        kv_seqlen = kv_seqlen.tolist()

    # fill attention_bias with BlockDiagonalMask
    with torch.no_grad():
        attn_bias = xops.fmha.BlockDiagonalMask.from_seqlens(
            q_seqlen=q_seqlen,
            kv_seqlen=kv_seqlen,
        )

    out = xops.memory_efficient_attention(
        query,
        key,
        value,
        attn_bias=attn_bias,
        p=dropout_p,
    )

    if rotate_value:
        out = apply_rotary_pos_emb(-q_pos_emb.unsqueeze(0), out)

    out = rearrange(out, "() n h d -> n (h d)")
    return out



# from pathlib import Path

# import pytest
# import torch
# import util
# from torch.utils.data import DataLoader

# from torch_brain.data import Dataset
# from torch_brain.data.sampler import SequentialFixedWindowSampler
# from torch_brain.data.collate import collate
# from torch_brain.models import POYOPlus, POYOPlusTokenizer
# from torch_brain.nn import InfiniteVocabEmbedding
# from brainsets_utils.taxonomy import decoder_registry
# from torch_brain.utils import move_to

# DATA_ROOT = Path(util.get_data_paths()["processed_dir"]) / "processed"


# def test_load_real_data():
#     ds = Dataset(DATA_ROOT, "train", [{"selection": [{"dandiset": "mc_maze_small"}]}])
#     sampling_intervals = ds.session_dict["mc_maze_small/jenkins_20090928_maze"][
#         "sampling_intervals"
#     ]
#     assert (
#         ds.get(
#             "mc_maze_small/jenkins_20090928_maze",
#             sampling_intervals.start[0],
#             sampling_intervals.end[0],
#         ).start
#         >= 0
#     )


# def test_collate_data():
#     unit_emb = InfiniteVocabEmbedding(2)
#     sess_emb = InfiniteVocabEmbedding(2)

#     tokenizer = POYOPlusTokenizer(
#         unit_tokenizer=unit_emb.tokenizer,
#         session_tokenizer=sess_emb.tokenizer,
#         decoder_registry=decoder_registry,
#         latent_step=1.0 / 8,
#         num_latents_per_step=4,
#         batch_type=["stacked", "stacked", "stacked"],
#     )

#     ds = Dataset(
#         DATA_ROOT,
#         "train",
#         [
#             {
#                 "selection": [
#                     {
#                         "dandiset": "mc_maze_small",
#                         "session": "jenkins_20090928_maze",
#                     }
#                 ],
#                 "config": {
#                     "multitask_readout": [
#                         {
#                             "decoder_id": "ARMVELOCITY2D",
#                             "weight": 2.0,
#                             "subtask_key": None,
#                         }
#                     ],
#                 },
#             }
#         ],
#         transform=tokenizer,
#     )

#     unit_emb.initialize_vocab(ds.unit_ids)
#     sess_emb.initialize_vocab(ds.session_ids)

#     sampler = SequentialFixedWindowSampler(
#         sampling_intervals=ds.get_sampling_intervals(),
#         window_length=1.0,
#     )
#     assert len(sampler) > 0

#     train_loader = DataLoader(ds, collate_fn=collate, batch_size=16, sampler=sampler)
#     for data in train_loader:
#         assert data["spike_timestamps"].shape[0] == 16
#         break


# # def test_collate_data_willett():
# #     print("test_collate_data_willett")
# #     ds = Dataset(
# #         DATA_ROOT,
# #         "train",
# #         [
# #             {
# #                 "selection": {
# #                     "dandiset": "willett_shenoy",
# #                     "sortset": "willett_shenoy_t5/t5.2019.05.08",
# #                 },
# #                 "metrics": [{"output_key": "WRITING_CHARACTER"}],
# #             }
# #         ],
# #     )
# #     assert len(ds) > 0

# #     od = OrderedDict({x: 1 for x in ds.unit_names})
# #     vocab = torchtext.vocab.vocab(od, specials=["NA"])

# #     collate_fn = Collate(
# #         num_latents_per_step=128,  # This was tied in train_poyo_1.py
# #         step=1.0 / 8,
# #         sequence_length=128,
# #         unit_vocab=vocab,
# #         decoder_registry=decoder_registry,
# #         weight_registry=weight_registry,
# #     )
# #     train_loader = DataLoader(
# #         ds, collate_fn=collate_fn, batch_size=4, drop_last=True, shuffle=True
# #     )
# #     for i, data in enumerate(train_loader):
# #         print(i)
# #         assert data["spike_timestamps"].shape[0] == 4
# #         npt.assert_allclose(
# #             data["output_weights"]["WRITING_CHARACTER"].detach().cpu().numpy(),
# #             1.0,
# #         )


# # def test_collate_data_perich():
# #     ds = Dataset(
# #         DATA_ROOT,
# #         "train",
# #         [
# #             {
# #                 "selection": {
# #                     "dandiset": "perich_miller",
# #                     "sortset": "chewie_20161013",
# #                 },
# #                 "metrics": [{"output_key": "CURSOR2D"}],
# #             }
# #         ],
# #     )
# #     assert len(ds) > 0

# #     od = OrderedDict({x: 1 for x in ds.unit_names})
# #     vocab = torchtext.vocab.vocab(od, specials=["NA"])

# #     collate_fn = Collate(
# #         num_latents_per_step=128,  # This was tied in train_poyo_1.py
# #         step=1.0 / 8,
# #         sequence_length=128,
# #         unit_vocab=vocab,
# #         decoder_registry=decoder_registry,
# #         weight_registry=weight_registry,
# #     )
# #     train_loader = DataLoader(ds, collate_fn=collate_fn, batch_size=16)
# #     for data in train_loader:
# #         assert data["spike_timestamps"].shape[0] == 16
# #         npt.assert_allclose(
# #             data["output_weights"]["CURSOR2D"].detach().cpu().numpy().max(),
# #             50.0,
# #         )
# #         npt.assert_allclose(
# #             data["output_weights"]["CURSOR2D"].detach().cpu().numpy().min(),
# #             1.0,
# #         )
# #         break


# def test_collated_data_model():
#     model = POYOPlus(
#         task_specs=decoder_registry,
#         backend_config="cpu",
#     )

#     tokenizer = POYOPlusTokenizer(
#         unit_tokenizer=model.unit_emb.tokenizer,
#         session_tokenizer=model.session_emb.tokenizer,
#         decoder_registry=decoder_registry,
#         latent_step=1.0 / 8,
#         num_latents_per_step=4,
#         batch_type=["stacked", "stacked", "stacked"],
#     )

#     ds = Dataset(
#         DATA_ROOT,
#         "train",
#         [
#             {
#                 "selection": [
#                     {
#                         "dandiset": "mc_maze_small",
#                         "session": "jenkins_20090928_maze",
#                     }
#                 ],
#                 "config": {
#                     "multitask_readout": [
#                         {
#                             "decoder_id": "ARMVELOCITY2D",
#                             "weight": 2.0,
#                             "subtask_key": None,
#                         }
#                     ],
#                 },
#             }
#         ],
#         transform=tokenizer,
#     )

#     model.unit_emb.initialize_vocab(ds.unit_ids)
#     model.session_emb.initialize_vocab(ds.session_ids)

#     sampler = SequentialFixedWindowSampler(
#         sampling_intervals=ds.get_sampling_intervals(),
#         window_length=1.0,
#     )
#     assert len(sampler) > 0

#     train_loader = DataLoader(ds, collate_fn=collate, batch_size=16, sampler=sampler)
#     device = "cpu"

#     model = model.to(device)
#     assert len(train_loader) > 0
#     for data in train_loader:
#         data = move_to(data, device)
#         output, loss, losses_taskwise = model(**data)

#         assert len(output) == 16
#         assert loss.shape == torch.Size([])  # it should be a scalar
#         assert len(losses_taskwise.keys()) == len(data["output_values"].keys())
#         break



import torch
import torch

from lightning.pytorch.callbacks import Callback


class UnitEmbeddingGradientRescaling(Callback):
    def __init__(self, dataset):
        super().__init__()

        session_weights = {"NA": 1.0}
        # session_info = dataset.session_info_dict[session_id]
        for session_id, session_info in dataset.session_info_dict.items():
            assert len(session_info["config"]["multitask_readout"]) == 1
            unit_weight = session_info["config"]["multitask_readout"][0]["weight"]
            session_weights[session_id] = 1.0 / unit_weight

        unit_ids = dataset.unit_ids
        session_ids = dataset.session_ids

        unit_weight = torch.ones(1 + len(unit_ids))
        for unit_idx, unit_id in enumerate(unit_ids):
            for session_idx, session_id in enumerate(session_ids):
                if session_id in unit_id:
                    unit_weight[unit_idx + 1] = session_weights[session_id]
                    break

        session_weight = torch.ones(1 + len(session_ids))
        for session_idx, session_id in enumerate(session_ids):
            session_weight[session_idx + 1] = session_weights[session_id]

        self.unit_weight = unit_weight
        self.session_weight = session_weight

    def on_after_backward(self, trainer, pl_module):
        # assert len(self.unit_weight) == len(pl_module.model.unit_emb.weight)

        pl_module.model.unit_emb.weight.grad.mul_(
            self.unit_weight.view(-1, 1).to(pl_module.model.unit_emb.weight.device)
        )
        pl_module.model.session_emb.weight.grad.mul_(
            self.session_weight.view(-1, 1).to(
                pl_module.model.session_emb.weight.device
            )
        )



import torch.nn as nn
import torch.nn.functional as F


class GEGLU(nn.Module):
    """Gated Gaussian Error Linear Unit (GEGLU) activation function, as introduced in
    the paper "GLU Variants Improve Transformer" (https://arxiv.org/abs/2002.05202).
    """

    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


class FeedForward(nn.Module):
    """A feed-forward network with GEGLU activation.

    Args:
        dim (int): Input and output dimension
        mult (int, optional): Multiplier for hidden dimension. Defaults to 4
        dropout (float, optional): Dropout probability. Defaults to 0.2
    """

    def __init__(self, dim, mult=4, dropout=0.2):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim * mult * 2),
            GEGLU(),
            nn.Dropout(p=dropout),
            nn.Linear(dim * mult, dim),
        )

    def forward(self, x):
        return self.net(x)



import torch


class RandomCrop:
    def __init__(self, crop_len):
        self.crop_len = crop_len

    def __call__(self, data):
        sequence_len = data.end - data.start

        if sequence_len <= self.crop_len:
            return data

        start = torch.rand(1).item() * (sequence_len - self.crop_len)
        end = start + self.crop_len

        return data.slice(start, end)



import math

import torch
from torch.optim.optimizer import Optimizer


class SparseLamb(Optimizer):
    r"""Implements the sparse variant of the Lamb algorithm.

    SparseLamb is a variant of the Lamb optimizer that only updates the parameters
    for which the gradient is non-zero. This is useful for models that do not always
    use all parameters during a single forward pass.

    By default, SparseLamb is equivalent to Lamb. To activate the sparse variant,
    set the `sparse` argument to `True`.

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps: term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay: weight decay (L2 penalty) (default: 0)
        clamp_value: clamp weight_norm in (0,clamp_value) (default: 10)
            set to a high value to avoid it (e.g 10e3)
        adam: always use trust ratio = 1, which turns this
            into Adam. Useful for comparison purposes. (default: False)
        debias: debias adam by (1 - beta**step) (default: False)
        sparse: only update the parameters that have non-zero gradients (default: False)

    __ https://arxiv.org/abs/1904.00962

    Note:
        Reference code: https://github.com/cybertronai/pytorch-lamb
    """

    def __init__(
        self,
        params,
        lr: float = 1e-3,
        betas=(0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0,
        clamp_value: float = 10,
        adam: bool = False,
        debias: bool = False,
        sparse: bool = False,
    ) -> None:
        if lr <= 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0.0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if weight_decay < 0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if clamp_value < 0.0:
            raise ValueError("Invalid clamp value: {}".format(clamp_value))

        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, sparse=sparse
        )
        self.clamp_value = clamp_value
        self.adam = adam
        self.debias = debias

        super(SparseLamb, self).__init__(params, defaults)

    def step(self, closure=None):
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad.data

                if grad.is_sparse:
                    msg = (
                        "Lamb does not support sparse gradients, "
                        "please consider SparseAdam instead"
                    )
                    raise RuntimeError(msg)

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1
                is_sparse = group["sparse"]
                if is_sparse:
                    # only update the values that are non-zero
                    mask = grad.abs().sum(dim=-1) > 0.0
                    is_sparse = not mask.all()

                if is_sparse:
                    mask_idx = mask.nonzero()
                    mask_idx = mask_idx.repeat((1, p.size(1)))

                    beta1_vec = torch.full_like(p, beta1)
                    beta1_vec[~mask] = 0
                    beta2_vec = torch.full_like(p, beta2)
                    beta2_vec[~mask] = 0

                    # Decay the first and second moment running average coefficient
                    # m_t
                    exp_avg.mul_(beta1_vec).scatter_add_(
                        0, mask_idx, grad[mask].mul_(1 - beta1)
                    )
                    # v_t
                    exp_avg_sq.mul_(beta2_vec).scatter_add_(
                        0, mask_idx, grad[mask].mul_(grad[mask]).mul_(1 - beta2)
                    )
                else:
                    # Decay the first and second moment running average coefficient
                    # m_t
                    exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                    # v_t
                    exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                # Paper v3 does not use debiasing.
                if self.debias:
                    raise NotImplementedError
                    bias_correction = math.sqrt(1 - beta2 ** state["step"])
                    bias_correction /= 1 - beta1 ** state["step"]
                else:
                    bias_correction = 1

                # Apply bias to lr to avoid broadcast.
                step_size = group["lr"] * bias_correction

                weight_norm = torch.norm(p.data).clamp(0, self.clamp_value)

                adam_step = exp_avg / exp_avg_sq.sqrt().add(group["eps"])
                if group["weight_decay"] != 0:
                    adam_step.add_(p.data, alpha=group["weight_decay"])

                adam_norm = torch.norm(adam_step)
                if weight_norm == 0 or adam_norm == 0:
                    trust_ratio = 1
                else:
                    trust_ratio = weight_norm / adam_norm
                state["weight_norm"] = weight_norm
                state["adam_norm"] = adam_norm
                state["trust_ratio"] = trust_ratio
                if self.adam:
                    trust_ratio = 1
                if is_sparse:
                    # alpha_vec = torch.full((p.size(0),), -step_size * trust_ratio)
                    p.data.scatter_add_(
                        0, mask_idx, adam_step[mask].mul_(-step_size * trust_ratio)
                    )
                else:
                    p.data.add_(adam_step, alpha=-step_size * trust_ratio)

        return loss



# from test_dataset_sim import (
#     dummy_data,
#     RUNNING_SPEED_MEAN,
#     RUNNING_SPEED_STD,
#     GABOR_POS_2D_MEAN,
#     GABOR_POS_2D_STD,
# )
# from torch_brain.data import Dataset
# from scripts.calculate_normalization_scales import calculate_zscales
# from brainsets_utils.taxonomy import decoder_registry
# from torch_brain.nn.multitask_readout import prepare_for_multitask_readout


# def make_dataset(dummy_data, data_config=None):
#     if not data_config:
#         data_config = [
#             {
#                 "selection": [
#                     {
#                         "dandiset": "allen_neuropixels_mock",
#                     }
#                 ],
#                 "config": {
#                     "multitask_readout": [
#                         {
#                             "decoder_id": "RUNNING_SPEED",
#                             "normalize_mean": RUNNING_SPEED_MEAN,
#                             "normalize_std": RUNNING_SPEED_STD,
#                         },
#                         {
#                             "decoder_id": "GABOR_POS_2D",
#                             "normalize_mean": [GABOR_POS_2D_MEAN, GABOR_POS_2D_MEAN],
#                             "normalize_std": [GABOR_POS_2D_STD, GABOR_POS_2D_STD],
#                         },
#                     ],
#                 },
#             }
#         ]
#     ds = Dataset(
#         dummy_data,
#         None,
#         data_config,
#     )
#     ds._check_for_data_leakage_flag = False

#     return ds


# def test_zscales_calculation(dummy_data):
#     ds = make_dataset(dummy_data)
#     zscales = calculate_zscales(ds)
#     assert round(zscales["RUNNING_SPEED"][0]) == RUNNING_SPEED_MEAN
#     assert round(zscales["RUNNING_SPEED"][1]) == RUNNING_SPEED_STD
#     assert round(zscales["GABOR_POS_2D"][0][0]) == GABOR_POS_2D_MEAN
#     assert round(zscales["GABOR_POS_2D"][0][1]) == GABOR_POS_2D_MEAN
#     assert round(zscales["GABOR_POS_2D"][1][0]) == GABOR_POS_2D_STD
#     assert round(zscales["GABOR_POS_2D"][1][1]) == GABOR_POS_2D_STD


# def test_zscaling_in_multitask_readout(dummy_data):
#     ds = make_dataset(dummy_data)
#     data = ds.get("allen_neuropixels_mock/20100101_01", 0, 1)
#     # prepare_for_multitask_readout should zscale the data if the config contains the mean and std
#     _, _, values_dict, _, _ = prepare_for_multitask_readout(data, decoder_registry)

#     for key, values in values_dict.items():
#         # should be standard normal distribution
#         assert round(values.mean()) == 0
#         assert round(values.std()) == 1



import copy
from collections import namedtuple
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import torch
from torch.utils.data._utils.collate import collate as _collate, default_collate_fn_map

import numpy as np


# pad
PaddedObject = namedtuple("PaddedObject", ["obj"])


def pad(obj):
    r"""Wrap an object to specify that it (or any of its members) should be padded to
    the maximum length in the batch. The object can be any of the objects that PyTorch's
    :obj:`default_collate` already supports.

    Args:
        obj: Can be tensors, numpy arrays, lists, tuples, or dictionaries.
    """
    return PaddedObject(obj)


def track_mask(input: Union[torch.Tensor, np.ndarray]):
    r"""Wrap an array or tensor to specify that its padding mask should be tracked.

    Args:
        input: An array or tensor.
    """
    return pad(torch.ones((len(input)), dtype=torch.bool))


def pad_collate_tensor_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    # todo this will be more optimal than any code we'll write? it's in C++
    return torch.nn.utils.rnn.pad_sequence(batch, batch_first=True, padding_value=0)


pad_collate_fn_map = copy.deepcopy(default_collate_fn_map)
pad_collate_fn_map[torch.Tensor] = pad_collate_tensor_fn


def pad_collate_object_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    return _collate([e.obj for e in batch], collate_fn_map=pad_collate_fn_map)


# pad8
Padded8Object = namedtuple("Padded8Object", ["obj"])


def pad8(obj):
    r"""Wrap an object to specify that it (or any of its members) should be padded to
    the maximum length in the batch. This function is similar to :obj:`pad` except that
    the padding length is rounded up to the nearest multiple of 8.

    Args:
        obj: Can be tensors, numpy arrays, lists, tuples, or dictionaries.
    """
    return Padded8Object(obj)


def track_mask8(input: Union[torch.Tensor, np.ndarray]):
    r"""Wrap an array or tensor to specify that its padding mask should be tracked. This
    is used in conjunction with :obj:`pad8`.

    Args:
        input: An array or tensor.
    """
    return pad8(torch.ones((len(input)), dtype=torch.bool))


def pad8_collate_tensor_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    max_len = max([elem.shape[0] for elem in batch])

    if max_len % 8 == 0:
        return pad_collate_tensor_fn(batch)

    elem = batch[0]
    batch.append(
        torch.zeros(
            (max_len + 8 - (max_len % 8), *elem.shape[1:]), dtype=batch[0].dtype
        )
    )

    return pad_collate_tensor_fn(batch)[:-1]


pad8_collate_fn_map = copy.deepcopy(default_collate_fn_map)
pad8_collate_fn_map[torch.Tensor] = pad8_collate_tensor_fn


def pad8_collate_object_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    return _collate([e.obj for e in batch], collate_fn_map=pad8_collate_fn_map)


# chain
ChainObject = namedtuple("ChainObject", ["obj", "allow_missing_keys"])
ChainBatchTrackerObject = namedtuple("ChainBatchTrackerObject", ["obj"])


def chain(obj, allow_missing_keys: bool = False):
    r"""Wrap an object to specify that it (or any of its members) should be stacked
    along the first dimension when batching. This approach is similar to PyTorch
    Geometric's collate approach for graphs. This function will chain all the sequences
    in the batch into one large sequence. To track which sample from the batch an
    element of the sequence came from use :func:`track_batch`.

    Args:
        obj: Can be tensors, numpy arrays, lists, tuples, or dictionaries.
        allow_missing_keys: If set to :obj:`True`, this allows the wrapped dictionary to
            have inconsistent keys across the batch. If an instance is missing keys,
            the corresponding values will be skipped when collating. Defaults to
            :obj:`False`.
    """
    if allow_missing_keys and not isinstance(obj, dict):
        raise TypeError(
            f"allow_missing_keys can only be used with dictionaries, got {type(obj)}."
        )
    return ChainObject(obj, allow_missing_keys)


def track_batch(input: Union[torch.Tensor, np.ndarray]):
    r"""Wrap an array or tensor to track the batch_index.

    Args:
        input: An array or tensor.
    """
    return ChainBatchTrackerObject(torch.ones((len(input)), dtype=torch.long))


def chain_collate_str_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    # No-op.
    return batch


def chain_collate_tensor_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    return torch.cat(batch, dim=0)


def chain_batch_tracker_collate_tensor_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    return _collate(
        [i * e.obj for i, e in enumerate(batch)],
        collate_fn_map=chain_collate_fn_map,
    )


chain_collate_fn_map = copy.deepcopy(default_collate_fn_map)
chain_collate_fn_map[str] = chain_collate_str_fn
chain_collate_fn_map[torch.Tensor] = chain_collate_tensor_fn
chain_collate_fn_map[ChainBatchTrackerObject] = chain_batch_tracker_collate_tensor_fn


def chain_collate_object_fn(
    batch,
    *,
    collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
):
    allow_missing_keys = batch[0].allow_missing_keys
    # check that flag is consistent
    if any((elem.allow_missing_keys != allow_missing_keys) for elem in batch):
        raise ValueError(
            "attribute 'allow_missing_keys' must be the same for all elements in the "
            "batch. Some elements allow missing keys while others do not."
        )

    if not allow_missing_keys:
        return _collate(
            [elem.obj for elem in batch], collate_fn_map=chain_collate_fn_map
        )
    else:
        unique_keys = set().union(*[elem.obj.keys() for elem in batch])
        return {
            key: _collate(
                [elem.obj[key] for elem in batch if key in elem.obj],
                collate_fn_map=chain_collate_fn_map,
            )
            for key in unique_keys
        }


# add all new types to collate fn map
# note that once a recipe is selected for a given object it cannot be overwritten
# this is to avoid the following example scenario where pad(obj) is called but obj a
# dict and one of its values was already wrapped in chain.
collate_fn_map = copy.deepcopy(default_collate_fn_map)
collate_fn_map[PaddedObject] = pad_collate_object_fn
collate_fn_map[Padded8Object] = pad8_collate_object_fn
collate_fn_map[ChainObject] = chain_collate_object_fn
collate_fn_map[ChainBatchTrackerObject] = chain_batch_tracker_collate_tensor_fn


def collate(batch):
    r"""Extension of PyTorch's :obj:`default_collate` function to enable more advanced
    collation of samples of variable lengths.

    To specify how the collation recipe, wrap the objects using :obj:`pad` or :obj:`chain`.
    If the wrapped object is an Iterable or Mapping, all its elements will inherite
    the collation recipe. All objects that are already supported by :obj:`default_collate`
    can be wrapped.

    If an object is not wrapped, the default collation recipe will be used. i.e. the
    outcome will be identical to :obj:`default_collate`.

    :obj:`pad` or :obj:`chain` do not track any padding masks or batch index, since that might not
    always be needed. Use :obj:`track_mask` or :obj:`track_batch` to track masks or batch index
    for a particular array or tensor.

    Args:
        batch: a single batch to be collated
    """
    return _collate(batch, collate_fn_map=collate_fn_map)



from .container import Compose, ConditionalChoice, RandomChoice
from .output_sampler import RandomOutputSampler
from .random_crop import RandomCrop
from .random_time_scaling import RandomTimeScaling
from .unit_dropout import TriangleDistribution, UnitDropout
from .unit_filter import UnitFilter, UnitFilterById



from abc import ABC, abstractmethod
from typing import Optional

import torch
import torch.nn.functional as F
from torchtyping import TensorType


class Loss(torch.nn.Module, ABC):
    r"""Base class for losses. All losses should support an optional weights argument."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        r"""Abstract method for computing the loss."""
        pass


class MSELoss(Loss):
    def forward(
        self,
        input: TensorType["batch_size", "dim"],
        target: TensorType["batch_size", "dim"],
        weights: Optional[TensorType["batch_size"]] = None,
    ) -> torch.Tensor:
        r"""Compute mean squared error loss.

        Args:
            input (Tensor): The input tensor.
            target (Tensor): The target tensor.
            weights (Tensor, optional): The weights tensor.
        """
        if input.ndim != 2:
            raise ValueError("Input must have 2 dimensions")
        if target.ndim != 2:
            raise ValueError("Target must have 2 dimensions")
        if weights is not None and weights.ndim != 1:
            raise ValueError("Weights must have 1 dimension")
        if weights is not None and input.shape[0] != weights.shape[0]:
            raise ValueError("Input and weights must have the same batch size")

        if weights is not None:
            loss_noreduce = F.mse_loss(input, target, reduction="none").mean(dim=1)
            return (weights * loss_noreduce).sum() / weights.sum()
        else:
            return F.mse_loss(input, target)


class CrossEntropyLoss(Loss):
    def forward(
        self,
        input: TensorType["batch_size", "dim"],
        target: TensorType["batch_size"],
        weights: Optional[TensorType["batch_size"]] = None,
    ) -> torch.Tensor:
        r"""Compute cross-entropy loss.

        Args:
            input (Tensor): The input tensor.
            target (Tensor): The target tensor.
            weights (Tensor, optional): The weights tensor.
        """
        if input.ndim != 2:
            raise ValueError("Input must have 2 dimensions")
        if target.ndim != 1:
            raise ValueError("Target must have 1 dimensions")
        if weights is not None and weights.ndim != 1:
            raise ValueError("Weights must have 1 dimension")
        if weights is not None and input.shape[0] != weights.shape[0]:
            raise ValueError("Input and weights must have the same batch size")

        if weights is not None:
            loss_noreduce = F.cross_entropy(input, target, reduction="none")
            return (weights * loss_noreduce).sum() / weights.sum()
        else:
            return F.cross_entropy(input, target)


class MallowDistanceLoss(Loss):
    def forward(
        self,
        input: TensorType["batch_size", "dim"],
        target: TensorType["batch_size"],
        weights: TensorType["batch_size"],
    ) -> torch.Tensor:
        r"""Compute Mallow distance loss.

        Args:
            input (Tensor): The input tensor.
            target (Tensor): The target tensor.
            weights (Tensor): The weights tensor.
        """
        num_classes = input.size(-1)
        input = torch.softmax(input, dim=-1).view(-1, num_classes)
        target = target.view(-1, 1)
        weights = weights.view(-1)
        # Mallow distance
        target = torch.zeros_like(input).scatter_(1, target, 1.0)
        # we compute the mallow distance as the sum of the squared differences
        loss = torch.mean(
            torch.square(torch.cumsum(target, dim=-1) - torch.cumsum(input, dim=-1)),
            dim=-1,
        )
        loss = (weights * loss).sum() / weights.sum()
        return loss



from enum import Enum
from typing import Dict, Tuple, Optional, Any, Callable

from pydantic.dataclasses import dataclass
import torch_brain


class DataType(Enum):
    """Enum defining the possible data types.

    Attributes:
        CONTINUOUS: For continuous-valued variables
        BINARY: For binary variables
        MULTINOMIAL: For multi-class variables
        MULTILABEL: For multi-label variables
    """

    CONTINUOUS = 0
    BINARY = 1
    MULTINOMIAL = 2
    MULTILABEL = 3


@dataclass
class ModalitySpec:
    """Specification for a modality.

    Attributes:
        dim: Dimension for this modality
        type: DataType enum specifying the data type
        loss_fn: Name of loss function to use for this modality
        timestamp_key: Key to access timestamps in the data object
        value_key: Key to access values in the data object
        id: Unique numeric ID assigned to this modality
    """

    id: int
    dim: int
    type: DataType
    timestamp_key: str  # can be overwritten
    value_key: str  # can be overwritten
    loss_fn: Callable  # can be overwritten


MODALITY_REGISTRY: Dict[str, ModalitySpec] = {}
_ID_TO_MODALITY: Dict[int, str] = {}


def register_modality(name: str, **kwargs: Any) -> int:
    """Register a new modality specification in the global registry.

    Args:
        name: Unique identifier for this modality
        **kwargs: Keyword arguments used to construct the ModalitySpec
            Must include: dim, type, loss_fn, timestamp_key, value_key

    Returns:
        int: Unique numeric ID assigned to this modality

    Raises:
        ValueError: If a modality with the given name already exists
    """
    # Check if modality already exists
    if name in MODALITY_REGISTRY:
        raise ValueError(f"Modality {name} already exists in registry")

    # Get next available ID
    next_id = len(MODALITY_REGISTRY) + 1

    # Create DecoderSpec from kwargs and set ID
    decoder_spec = ModalitySpec(**kwargs, id=next_id)

    # Add to registries
    MODALITY_REGISTRY[name] = decoder_spec
    _ID_TO_MODALITY[next_id] = name

    return next_id


def get_modality_by_id(modality_id: int) -> ModalitySpec:
    """Get a modality specification by its ID.

    Args:
        modality_id: The numeric ID of the modality to retrieve

    Returns:
        ModalitySpec: The modality specification

    Raises:
        KeyError: If no modality exists with the given ID
    """
    if modality_id not in _ID_TO_MODALITY:
        raise KeyError(f"No modality found with ID {modality_id}")
    return _ID_TO_MODALITY[modality_id]


register_modality(
    "cursor_velocity_2d",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="cursor.timestamps",
    value_key="cursor.vel",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "cursor_position_2d",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="cursor.timestamps",
    value_key="cursor.pos",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "arm_velocity_2d",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="behavior.timestamps",
    value_key="behavior.hand_vel",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "drifting_gratings_orientation",
    dim=8,
    type=DataType.MULTINOMIAL,
    timestamp_key="drifting_gratings.timestamps",
    value_key="drifting_gratings.orientation_id",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "drifting_gratings_temporal_frequency",
    dim=5,  # [1,2,4,8,15]
    type=DataType.MULTINOMIAL,
    timestamp_key="drifting_gratings.timestamps",
    value_key="drifting_gratings.temporal_frequency_id",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "natural_movie_one_frame",
    dim=900,
    type=DataType.MULTINOMIAL,
    timestamp_key="natural_movie_one.timestamps",
    value_key="natural_movie_one.frame",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "natural_movie_two_frame",
    dim=900,
    type=DataType.MULTINOMIAL,
    timestamp_key="natural_movie_two.timestamps",
    value_key="natural_movie_two.frame",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "natural_movie_three_frame",
    dim=3600,
    type=DataType.MULTINOMIAL,
    timestamp_key="natural_movie_three.timestamps",
    value_key="natural_movie_three.frame",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "locally_sparse_noise_frame",
    dim=8000,
    type=DataType.MULTINOMIAL,
    timestamp_key="locally_sparse_noise.timestamps",
    value_key="locally_sparse_noise.frame",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "static_gratings_orientation",
    dim=6,
    type=DataType.MULTINOMIAL,
    timestamp_key="static_gratings.timestamps",
    value_key="static_gratings.orientation_id",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "static_gratings_spatial_frequency",
    dim=5,
    type=DataType.MULTINOMIAL,
    timestamp_key="static_gratings.timestamps",
    value_key="static_gratings.spatial_frequency_id",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "static_gratings_phase",
    dim=5,
    type=DataType.MULTINOMIAL,
    timestamp_key="static_gratings.timestamps",
    value_key="static_gratings.phase_id",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "natural_scenes",
    dim=119,  # image classes [0,...,118]
    type=DataType.MULTINOMIAL,
    timestamp_key="natural_scenes.timestamps",
    value_key="natural_scenes.frame",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "gabor_orientation",
    dim=4,  # [0, 1, 2, 3]
    type=DataType.MULTINOMIAL,
    timestamp_key="gabors.timestamps",
    value_key="gabors.gabors_orientation",
    loss_fn=torch_brain.nn.loss.CrossEntropyLoss(),
)

register_modality(
    "gabor_pos_2d",  # 9x9 grid modeled as (x, y) coordinates
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="gabors.timestamps",
    value_key="gabors.pos_2d",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "running_speed",
    dim=1,
    type=DataType.CONTINUOUS,
    timestamp_key="running.timestamps",
    value_key="running.running_speed",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "gaze_pos_2d",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="gaze.timestamps",
    value_key="gaze.pos_2d",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "pupil_location",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="pupil.timestamps",
    value_key="pupil.location",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)

register_modality(
    "pupil_size_2d",
    dim=2,
    type=DataType.CONTINUOUS,
    timestamp_key="pupil.timestamps",
    value_key="pupil.size_2d",
    loss_fn=torch_brain.nn.loss.MSELoss(),
)



import pytest
import torch
import lightning as L
import torchmetrics
from collections import defaultdict

from torch_brain.utils.stitcher import MultiTaskDecodingStitchEvaluator


class MockDataModule:
    def __init__(self, val_sequence_index, test_sequence_index):
        self.val_sequence_index = val_sequence_index
        self.test_sequence_index = test_sequence_index


class MockTrainer:
    def __init__(self, val_sequence_index, test_sequence_index):
        self.datamodule = MockDataModule(val_sequence_index, test_sequence_index)
        self.loggers = []
        self.is_global_zero = True


@pytest.fixture
def mock_metrics():
    return {
        "session1": {
            "task1": {
                "accuracy": torchmetrics.Accuracy(task="multiclass", num_classes=3),
                "f1": torchmetrics.F1Score(task="multiclass", num_classes=3),
            },
            "task2": {"mse": torchmetrics.MeanSquaredError()},
        }
    }


@pytest.fixture
def evaluator(mock_metrics):
    return MultiTaskDecodingStitchEvaluator(metrics=mock_metrics)


def test_initialization(evaluator):
    assert evaluator.metrics is not None
    assert "session1" in evaluator.metrics
    assert "task1" in evaluator.metrics["session1"]
    assert "task2" in evaluator.metrics["session1"]


def test_on_validation_epoch_start(evaluator):
    val_sequence_index = torch.tensor([0, 0, 1, 1, 1])
    test_sequence_index = torch.tensor([0, 1, 1, 2, 2])  # different sequence
    trainer = MockTrainer(val_sequence_index, test_sequence_index)

    evaluator.on_validation_epoch_start(trainer, None)

    assert evaluator.sample_ptr == 0
    assert len(evaluator.cache) == 2  # max val_sequence_index + 1
    assert evaluator.counter == [0, 0]
    assert torch.equal(evaluator.cache_flush_threshold, torch.tensor([2, 3]))


def test_on_test_epoch_start(evaluator):
    val_sequence_index = torch.tensor([0, 0, 1, 1, 1])
    test_sequence_index = torch.tensor([0, 1, 1, 2, 2])  # different sequence
    trainer = MockTrainer(val_sequence_index, test_sequence_index)

    evaluator.on_test_epoch_start(trainer, None)

    assert evaluator.sample_ptr == 0
    assert len(evaluator.cache) == 3  # max test_sequence_index + 1
    assert evaluator.counter == [0, 0, 0]  # three counters for three sequences
    assert torch.equal(evaluator.cache_flush_threshold, torch.tensor([1, 2, 2]))



import logging

import numpy as np
import pytest

from temporaldata import ArrayDict, Data, IrregularTimeSeries
from torch_brain.transforms.unit_filter import UnitFilter, UnitFilterById

logger = logging.getLogger(__name__)


@pytest.fixture
def mock_data():
    timestamps = np.arange(200)
    unit_index = [0] * 10 + [1] * 20 + [2] * 70 + [3] * 10 + [4] * 20 + [5] * 70
    unit_index = np.array(unit_index)
    types = np.zeros(200)
    data = Data(
        spikes=IrregularTimeSeries(
            timestamps=timestamps,
            unit_index=unit_index,
            types=types,
            domain="auto",
        ),
        units=ArrayDict(
            id=np.array(
                [
                    "sorted_a",
                    "unsorted_a",
                    "sorted_b",
                    "unsorted_b",
                    "sorted_c",
                    "unsorted_c",
                ]
            ),
        ),
        domain="auto",
    )
    return data


def test_unit_filter_w_keyword(mock_data):
    transform = UnitFilterById(
        "unsorted", field="spikes", reset_index=True, keep_matches=True
    )
    data_t = transform(mock_data)

    expected_unit_ids = ["unsorted_a", "unsorted_b", "unsorted_c"]
    assert np.array_equal(data_t.units.id, expected_unit_ids)

    expected_unit_index = [0] * 20 + [1] * 10 + [2] * 70
    assert np.array_equal(data_t.spikes.unit_index, expected_unit_index)

    expected_timestamps = np.concatenate(
        [np.arange(10, 30), np.arange(100, 110), np.arange(130, 200)]
    )
    assert np.array_equal(data_t.spikes.timestamps, expected_timestamps)


def test_unit_filter_w_regex(mock_data):
    transform = UnitFilterById(
        r"^sorted_.*", field="spikes", reset_index=True, keep_matches=True
    )
    data_t = transform(mock_data)

    expected_unit_ids = ["sorted_a", "sorted_b", "sorted_c"]
    assert np.array_equal(data_t.units.id, expected_unit_ids)

    expected_unit_index = [0] * 10 + [1] * 70 + [2] * 20
    assert np.array_equal(data_t.spikes.unit_index, expected_unit_index)

    expected_timestamps = np.concatenate(
        [np.arange(10), np.arange(30, 100), np.arange(110, 130)]
    )
    assert np.array_equal(data_t.spikes.timestamps, expected_timestamps)



import numpy as np

from temporaldata import (
    Data,
    ArrayDict,
    Interval,
    IrregularTimeSeries,
    RegularTimeSeries,
)
from torch_brain.transforms.random_time_scaling import RandomTimeScaling
from torch_brain.transforms.unit_dropout import TriangleDistribution, UnitDropout


def test_irregular_scaling():
    timestamps = np.linspace(0, 100, 100)
    unit_index = [0] * 10 + [1] * 20 + [2] * 70
    unit_index = np.array(unit_index)

    # shuffle units
    np.random.shuffle(unit_index)
    types = np.zeros(100)

    data = Data(
        spikes=IrregularTimeSeries(
            timestamps=timestamps,
            unit_index=unit_index,
            types=types,
            domain="auto",
        ),
        units=ArrayDict(
            id=np.array(["a", "b", "c"]),
        ),
        domain=Interval(start=np.array([0]), end=np.array([100])),
    )

    rts = RandomTimeScaling(min_scale=2.0, max_scale=2.0)
    data_prime = rts(data)

    assert data_prime.start == 0
    assert data_prime.end == 200
    assert data_prime.spikes.timestamps[0] == 0
    assert data_prime.spikes.timestamps[-1] == 200

    rts = RandomTimeScaling(
        min_scale=1.0, max_scale=1.0, min_offset=-1.0, max_offset=-1.0
    )
    data_prime = rts(data)

    assert data_prime.start == -1
    assert data_prime.end == 99
    assert data_prime.spikes.timestamps[0] == -1
    assert data_prime.spikes.timestamps[-1] == 99
    assert data_prime.domain.start == -1
    assert data_prime.domain.end == 99
    assert data_prime.spikes.domain.start == -1
    assert data_prime.spikes.domain.end == 99


def test_regular_scaling():
    lfps = RegularTimeSeries(
        sampling_rate=10.0,
        data=np.random.randn(100),
        domain="auto",
    )
    data = Data(lfps=lfps, domain=lfps.domain)

    rts = RandomTimeScaling(min_scale=2.0, max_scale=2.0)
    data_prime = rts(data)

    assert data_prime.lfps.domain.start == 0
    assert data_prime.lfps.domain.end == 19.8



import numpy as np
import pytest
from temporaldata import Data, Interval

from torch_brain.utils.weights import resolve_weights_based_on_interval_membership


def MockData():
    data = Data(
        movement_periods=Data(
            reach_period=Interval(start=np.array([2.0]), end=np.array([4.0])),
            hold_period=Interval(start=np.array([3.0]), end=np.array([5.0])),
            domain="auto",
        ),
        cursor_outlier_segments=Interval(start=np.array([3.5]), end=np.array([4.5])),
        empty_interval=Interval(start=np.array([]), end=np.array([])),
        out_of_bounds_interval=Interval(start=np.array([6.0]), end=np.array([7.0])),
        invalid_attribute="not_an_interval",
        domain=Interval(start=np.array([0.0]), end=np.array([10.0])),
    )

    return data


def test_default_weights():
    """Test that default weights (no config) are all ones."""
    timestamps = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    data = MockData()

    weights = resolve_weights_based_on_interval_membership(timestamps, data)
    assert np.all(weights == 1.0)
    assert weights.shape == timestamps.shape


def test_single_interval_weights():
    """Test weights with a single interval."""
    timestamps = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    data = MockData()
    config = {"movement_periods.reach_period": 5.0}

    weights = resolve_weights_based_on_interval_membership(timestamps, data, config)
    expected = np.array([1.0, 5.0, 5.0, 1.0, 1.0])
    np.testing.assert_array_equal(weights, expected)


def test_overlapping_intervals():
    """Test weights with overlapping intervals."""
    timestamps = np.array([1.0, 2.0, 3.0, 3.75, 4.75])
    data = MockData()
    config = {
        "movement_periods.reach_period": 2.0,
        "movement_periods.hold_period": 3.0,
        "cursor_outlier_segments": 0.0,
    }

    weights = resolve_weights_based_on_interval_membership(timestamps, data, config)
    # At t=3.75, all three intervals overlap: 2.0 * 3.0 * 0.0 = 0.0
    expected = np.array([1.0, 2.0, 6.0, 0.0, 3.0])
    np.testing.assert_array_equal(weights, expected)


def test_empty_interval():
    """Test weights with an empty interval."""
    timestamps = np.array([1.0, 2.0, 3.0])
    data = MockData()
    config = {"empty_interval": 2.0}

    weights = resolve_weights_based_on_interval_membership(timestamps, data, config)
    expected = np.array([1.0, 1.0, 1.0])
    np.testing.assert_array_equal(weights, expected)


def test_out_of_bounds_interval():
    """Test weights with an out of bounds interval."""
    timestamps = np.array([1.0, 2.0, 3.0])
    data = MockData()
    config = {"out_of_bounds_interval": 2.0}

    weights = resolve_weights_based_on_interval_membership(timestamps, data, config)
    expected = np.array([1.0, 1.0, 1.0])
    np.testing.assert_array_equal(weights, expected)


def test_invalid_interval_type():
    """Test that an error is raised for invalid interval type."""
    timestamps = np.array([1.0, 2.0, 3.0])
    data = MockData()
    config = {"invalid_attribute": 2.0}

    with pytest.raises(ValueError, match="Expected an Interval object"):
        resolve_weights_based_on_interval_membership(timestamps, data, config)



import time
import subprocess
import logging

try:
    import lightning as L

    LIGHTNING_AVAILABLE = True
except ImportError:
    LIGHTNING_AVAILABLE = False

    class DummyCallback:
        pass

    L = type("L", (), {"Callback": DummyCallback})


def _check_lightning_available(cls):
    """Raise an error if Lightning is not available."""
    if not LIGHTNING_AVAILABLE:
        raise ImportError(
            f"Lightning is not installed. Please install it with "
            f"`pip install lightning~=2.3` to use {cls.__name__}."
        )


class EpochTimeLogger(L.Callback):
    r"""Lightning callback to log the time taken for each epoch.
    Args:
        enable (bool, optional): Whether to enable the callback. Defaults to `True`.
    """

    def __init__(self, enable=True):
        _check_lightning_available(self.__class__)
        self.enable = enable

    def on_train_epoch_start(self, trainer, pl_module):
        if self.enable:
            self.start_time = time.time()

    def on_train_epoch_end(self, trainer, pl_module):
        if self.enable:
            epoch_time = time.time() - self.start_time
            pl_module.log("epoch_time", epoch_time, sync_dist=True)


class ModelWeightStatsLogger(L.Callback):
    r"""Lightning callback to log the mean and standard deviation of the weights and
    gradients of the model.
    Args:
        enable (bool, optional): Whether to enable the callback. Defaults to `True`.
        grads (bool, optional): Whether to log the statistics of gradients.
            Defaults to `True`.
        module_name (str, optional): The name of the :class:`torch.nn.Module` object inside your
            :class:`lightning.LightningModule` object to log the statistics of. Defaults to "model".
            This name is prefixed to the keys in the logs, allowing logging the statistics
            of multiple modules with multiple instances of this callback.
    """

    def __init__(self, enable=True, grads=True, module_name="model", every_n_epoch=1):
        _check_lightning_available(self.__class__)
        self.enable = enable
        self.grads = grads
        self.module_name = module_name
        self.every_n_epoch = every_n_epoch

    def on_train_epoch_end(self, trainer, pl_module):
        if self.enable:
            if pl_module.current_epoch % self.every_n_epoch != 0:
                return

            model = getattr(pl_module, self.module_name)
            for tag, value in model.named_parameters():
                pl_module.log(
                    f"{self.module_name}_weights/mean_{tag}",
                    value.mean(),
                    sync_dist=True,
                )
                if len(value) > 1:
                    pl_module.log(
                        f"{self.module_name}_weights/std_{tag}",
                        value.std(),
                        sync_dist=True,
                    )
                if self.grads and value.grad is not None:
                    pl_module.log(
                        f"{self.module_name}_grads/mean_{tag}",
                        value.grad.mean(),
                        sync_dist=True,
                    )


class MemInfo(L.Callback):
    r"""Lightning callback to print the memory information of the system at the
    beginning of the training. This uses the `cat /proc/meminfo` command to get the
    memory information.
    """

    def on_train_start(self, trainer, pl_module):
        _check_lightning_available(self.__class__)
        # Log the output of `cat /proc/meminfo` using a shell script.
        try:
            # Execute the command and capture its output
            result = subprocess.run(
                ["cat", "/proc/meminfo"],
                capture_output=True,
                text=True,
                check=True,
            )
            result = result.stdout
        except subprocess.CalledProcessError as e:
            print(f"Command failed with error: {e}")
            result = ""

        # Log the output
        logging.info(f"Memory info: \n{result}")



import copy
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import h5py
import numpy as np
import omegaconf
import torch
from temporaldata import Data, Interval


@dataclass
class DatasetIndex:
    r"""The dataset can be indexed by specifying a recording id and a start and end time."""

    recording_id: str
    start: float
    end: float


default_session_id_prefix_fn = lambda data: f"{data.brainset.id}/"
default_unit_id_prefix_fn = lambda data: f"{data.brainset.id}/{data.session.id}/"
default_subject_id_prefix_fn = lambda data: f"{data.brainset.id}/"


class Dataset(torch.utils.data.Dataset):
    r"""This class abstracts a collection of lazily-loaded Data objects. Each data object
    corresponds to a full recording. It is never fully loaded into memory, but rather
    lazy-loaded on-the-fly from disk.

    The dataset can be indexed by a recording id and a start and end times using the `get`
    method. This definition is a deviation from the standard PyTorch Dataset definition,
    which generally presents the dataset directly as samples.
    In this case, the Dataset by itself does not provide you with samples, but rather the
    means to flexibly work and access complete recordings.

    Within this framework, it is the job of the sampler to provide a list of
    :class:`DatasetIndex` objects that are used to slice the dataset into samples (see
    Samplers).

    The lazy loading is done both in:
        - time: only the requested time interval is loaded, without having to load the entire
          recording into memory, and
        - attributes: attributes are not loaded until they are requested, this is useful when
          only a small subset of the attributes are actually needed.

    References to the underlying hdf5 files will be opened, and will only be closed when
    the Dataset object is destroyed.

    Args:
        root: The root directory of the dataset.
        config: The configuration file specifying the sessions to include.
        brainset: The brainset to include. This is used to specify a single brainset,
            and can only be used if config is not provided.
        session: The session to include. This is used to specify a single session, and
            can only be used if config is not provided.
        split: The split of the dataset. This is used to determine the sampling intervals
            for each session. The split is optional, and is used to load a subset of the data
            in a session based on a predefined split.
        transform: A transform to apply to the data. This transform should be a callable
            that takes a Data object and returns a Data object.
        unit_id_prefix_fn:
            A function to generate prefix strings for unit IDs to ensure uniqueness across
            the dataset. It takes a Data object as input and returns a string that would be
            prefixed to all unit ids in that Data object.
            Default corresponds to the function `lambda data: f"{data.brainset.id}/{data.session.id}/"`
        session_id_prefix_fn: Same as unit_id_prefix_fn but for session ids.
            Default corresponds to the function `lambda data: f"{data.brainset.id}/"`
        subject_id_prefix_fn: Same as unit_id_prefix_fn but for subject ids.
            Default corresponds to the function `lambda data: f"{data.brainset.id}/"`
    """

    _check_for_data_leakage_flag: bool = True
    _open_files: Optional[Dict[str, h5py.File]] = None
    _data_objects: Optional[Dict[str, Data]] = None

    def __init__(
        self,
        root: str,
        *,
        config: Optional[str] = None,
        recording_id: Optional[str] = None,
        split: Optional[str] = None,
        transform: Optional[Callable[[Data], Any]] = None,
        unit_id_prefix_fn: Callable[[Data], str] = default_unit_id_prefix_fn,
        session_id_prefix_fn: Callable[[Data], str] = default_session_id_prefix_fn,
        subject_id_prefix_fn: Callable[[Data], str] = default_subject_id_prefix_fn,
    ):
        super().__init__()
        self.root = root
        self.config = config
        self.split = split
        self.transform = transform
        self.unit_id_prefix_fn = unit_id_prefix_fn
        self.session_id_prefix_fn = session_id_prefix_fn
        self.subject_id_prefix_fn = subject_id_prefix_fn

        if config is not None:
            assert (
                recording_id is None
            ), "Cannot specify recording_id when using config."

            if isinstance(config, omegaconf.listconfig.ListConfig):
                config = omegaconf.OmegaConf.to_container(config)
            elif Path(config).is_file():
                config = omegaconf.OmegaConf.load(config)
            else:
                raise ValueError(f"Could not open configuration file: '{config}'")

            self.recording_dict = self._look_for_files(config)

        elif recording_id is not None:
            self.recording_dict = {
                recording_id: {
                    "filename": Path(self.root) / (recording_id + ".h5"),
                    "config": {},
                }
            }
        else:
            raise ValueError("Please either specify a config file or a recording_id.")

        self._open_files = {
            recording_id: h5py.File(recording_info["filename"], "r")
            for recording_id, recording_info in self.recording_dict.items()
        }

        self._data_objects = {
            recording_id: Data.from_hdf5(f, lazy=True)
            for recording_id, f in self._open_files.items()
        }

    def _close_open_files(self):
        """Closes the open files and deletes open data objects.
        This is useful when you are done with the dataset.
        """
        if self._open_files is not None:
            for f in self._open_files.values():
                f.close()
            self._open_files = None

        self._data_objects = None

    def __del__(self):
        self._close_open_files()

    def _look_for_files(self, config: omegaconf.DictConfig) -> Dict[str, Dict]:
        recording_dict = {}

        for i, selection_list in enumerate(config):
            selection = selection_list["selection"]

            # parse selection
            if len(selection) == 0:
                raise ValueError(
                    f"Selection {i} is empty. Please at least specify a brainset."
                )

            for subselection in selection:
                if subselection.get("brainset", "") == "":
                    raise ValueError(f"Please specify a brainset to include.")

                # Get a list of all the potentially chunks in this dataset.
                brainset_dir = Path(self.root) / subselection["brainset"]
                files = list(brainset_dir.glob("*.h5"))
                session_ids = sorted([f.stem for f in files])

                if len(session_ids) == 0:
                    raise ValueError(
                        f"No files found in {brainset_dir}. This is either a problem "
                        "with the provided path or the dataset has not been downloaded "
                        "and/or processed. For supported brainsets, please refer to "
                        "https://github.com/neuro-galaxy/brainsets"
                    )

                # Perform selection. Right now, we are limiting ourselves to session,
                # and subject filters, but we could make selection more flexible in the
                # future.
                sel_session = subselection.get("session", None)
                sel_sessions = subselection.get("sessions", None)
                sel_subject = subselection.get("subject", None)
                sel_subjects = subselection.get("subjects", None)
                # exclude_sessions allows you to exclude some sessions from the selection.
                # example use: you want to train on the complete brainset, but leave out
                # a few sessions for evaluating transfer performance.
                sel_exclude_sessions = subselection.get("exclude_sessions", None)

                # if subject is used for selection, we need to load all the files and
                # extract the subjects ids
                if sel_subject is not None or sel_subjects is not None:
                    all_session_subjects = []
                    for session_id in session_ids:
                        with h5py.File(brainset_dir / (session_id + ".h5"), "r") as f:
                            session_data = Data.from_hdf5(f, lazy=True)
                            all_session_subjects.append(session_data.subject.id)

                filtered = False
                if sel_session is not None:
                    assert (
                        sel_session in session_ids
                    ), f"Session {sel_session} not found in brainset {subselection['brainset']}"
                    session_ids = [sel_session]
                    filtered = True

                if sel_sessions is not None:
                    assert (
                        not filtered
                    ), "Cannot specify session AND sessions in selection"

                    # Check that all sortsets are in the brainset.
                    for session in sel_sessions:
                        assert (
                            session in session_ids
                        ), f"Session {session} not found in brainset {subselection['brainset']}"

                    session_ids = sorted(sel_sessions)
                    filtered = True

                if sel_subject is not None:
                    assert (
                        not filtered
                    ), "Cannot specify subject AND session(s) in selection"

                    assert (
                        sel_subject in all_session_subjects
                    ), f"Could not find subject {sel_subject} in brainset {subselection['brainset']}"

                    session_ids = [
                        session
                        for i, session in enumerate(session_ids)
                        if all_session_subjects[i] == sel_subject
                    ]
                    filtered = True

                if sel_subjects is not None:
                    assert (
                        not filtered
                    ), "Cannot specify subjects AND subject/session(s) in selection"

                    # Make sure all subjects asked for are in the brainset
                    sel_subjects = set(sel_subjects)
                    assert sel_subjects.issubset(all_session_subjects), (
                        f"Could not find subject(s) {sel_subjects - all_session_subjects} "
                        f" in brainset {subselection['brainset']}"
                    )

                    session_ids = [
                        session
                        for i, session in enumerate(session_ids)
                        if all_session_subjects[i] in sel_subjects
                    ]
                    filtered = True

                # Exclude sortsets if asked.
                if sel_exclude_sessions is not None:
                    session_ids = [
                        session
                        for session in session_ids
                        if session not in sel_exclude_sessions
                    ]

                assert (
                    len(session_ids) > 0
                ), f"No sessions left after filtering for selection {subselection['brainset']}"

                # Now we get the session-level information
                config = selection_list.get("config", {})

                for session_id in session_ids:
                    recording_id = subselection["brainset"] + "/" + session_id

                    if recording_id in recording_dict:
                        raise ValueError(
                            f"Recording {recording_id} is already included in the dataset."
                            "Please verify that it is only selected once."
                        )

                    recording_dict[recording_id] = dict(
                        filename=(Path(self.root) / (recording_id + ".h5")),
                        config=config,
                    )

        return recording_dict

    def get(self, recording_id: str, start: float, end: float):
        r"""This is the main method to extract a slice from a recording. It returns a
        Data object that contains all data for recording :obj:`recording_id` between
        times :obj:`start` and :obj:`end`.

        Args:
            recording_id: The recording id of the slice. This is usually
                <brainset_id>/<session_id>
            start: The start time of the slice.
            end: The end time of the slice.
        """
        data = copy.copy(self._data_objects[recording_id])
        # TODO: add more tests to make sure that slice does not modify the original data object
        # note there should be no issues as long as the self._data_objects stay lazy
        sample = data.slice(start, end)

        if self._check_for_data_leakage_flag and self.split is not None:
            sample._check_for_data_leakage(self.split)

        self._update_data_with_prefixed_ids(sample)
        sample.config = self.recording_dict[recording_id]["config"]

        return sample

    def get_recording_data(self, recording_id: str):
        r"""Returns the data object corresponding to the recording :obj:`recording_id`.
        If the split is not :obj:`None`, the data object is sliced to the allowed sampling
        intervals for the split, to avoid any data leakage. :obj:`RegularTimeSeries`
        objects are converted to :obj:`IrregularTimeSeries` objects, since they are
        most likely no longer contiguous.

        .. warning::
            This method might load the full data object in memory, avoid multiple calls
            to this method if possible.
        """
        data = copy.copy(self._data_objects[recording_id])

        # get allowed sampling intervals
        if self.split is not None:
            sampling_intervals = self.get_sampling_intervals()[recording_id]
            data = data.select_by_interval(sampling_intervals)
            if self._check_for_data_leakage_flag:
                data._check_for_data_leakage(self.split)
        else:
            data = copy.deepcopy(data)

        self._update_data_with_prefixed_ids(data)
        return data

    def get_sampling_intervals(self):
        r"""Returns a dictionary of sampling intervals for each session.
        This represents the intervals that can be sampled from each session.

        Note that these intervals will change depending on the split. If no split is
        provided, the full domain of the data is used.
        """
        sampling_intervals_dict = {}
        for recording_id in self.recording_dict.keys():
            sampling_domain = (
                f"{self.split}_domain" if self.split is not None else "domain"
            )
            sampling_intervals = getattr(
                self._data_objects[recording_id], sampling_domain
            )
            sampling_intervals_modifier_code = self.recording_dict[recording_id][
                "config"
            ].get("sampling_intervals_modifier", None)
            if sampling_intervals_modifier_code is not None:
                local_vars = {
                    "data": copy.deepcopy(self._data_objects[recording_id]),
                    "sampling_intervals": sampling_intervals,
                    "split": self.split,
                }
                try:
                    exec(sampling_intervals_modifier_code, {}, local_vars)
                except NameError as e:
                    error_message = (
                        f"{e}. Variables that are passed to the sampling_intervals_modifier "
                        f"are: {list(local_vars.keys())}"
                    )
                    raise NameError(error_message) from e
                except Exception as e:
                    error_message = (
                        f"Error while executing sampling_intervals_modifier defined in "
                        f"the config file for session {recording_id}: {e}"
                    )
                    raise type(e)(error_message) from e

                sampling_intervals = local_vars.get("sampling_intervals")
            sampling_intervals_dict[recording_id] = sampling_intervals
        return sampling_intervals_dict

    def get_recording_config_dict(self):
        r"""Returns configs for each session in the dataset as a dictionary."""
        ans = {}
        for recording_id in self.recording_dict.keys():
            ans[recording_id] = self.recording_dict[recording_id]["config"]
        return ans

    def _get_unit_ids_with_prefix(self, data: Data) -> np.ndarray:
        r"""Return unit ids with prefix applied"""
        prefix_str = self.unit_id_prefix_fn(data)
        # Check numpy version and use appropriate function
        if np.__version__ >= "2.0":
            return np.strings.add(prefix_str, data.units.id.astype(str))
        else:
            return np.core.defchararray.add(prefix_str, data.units.id.astype(str))

    def _get_session_id_with_prefix(self, data: Data) -> str:
        r"""Return session id with prefix applied"""
        return f"{self.session_id_prefix_fn(data)}{data.session.id}"

    def _get_subject_id_with_prefix(self, data: Data) -> str:
        r"""Return subject id with prefix applied"""
        return f"{self.subject_id_prefix_fn(data)}{data.subject.id}"

    def _update_data_with_prefixed_ids(self, data: Data):
        r"""Inplace add prefixes to unit ids, session id, and subect id"""
        if hasattr(data, "units"):
            data.units.id = self._get_unit_ids_with_prefix(data)

        if hasattr(data, "session"):
            data.session.id = self._get_session_id_with_prefix(data)

        if hasattr(data, "subject"):
            data.subject.id = self._get_subject_id_with_prefix(data)

    def get_unit_ids(self):
        r"""Returns all unit ids in the dataset."""
        unit_ids_list = []
        for data in self._data_objects.values():
            unit_ids = self._get_unit_ids_with_prefix(data)
            unit_ids_list.extend(unit_ids)
        return unit_ids_list

    def get_session_ids(self):
        r"""Returns the session ids of the dataset."""
        ans = []
        for data in self._data_objects.values():
            ans.append(self._get_session_id_with_prefix(data))
        return sorted(ans)

    def get_subject_ids(self):
        r"""Returns all subject ids in the dataset."""
        subject_ids = []
        for data in self._data_objects.values():
            subject_ids.append(self._get_subject_id_with_prefix(data))
        return sorted(list(set(subject_ids)))

    def get_brainset_ids(self):
        r"""Returns all brainset ids in the dataset."""
        brainset_ids = []
        for data in self._data_objects.values():
            brainset_ids.append(data.brainset.id)
        return sorted(list(set(brainset_ids)))

    def disable_data_leakage_check(self):
        r"""Disables the data leakage check.

        .. warning::
            Only do this you are absolutely sure that there is no leakage between the
            current split and other splits (eg. the test split).
        """
        self._check_for_data_leakage_flag = False
        logging.warn(
            f"Data leakage check is disabled. Please be absolutely sure that there is "
            f"no leakage between {self.split} and other splits."
        )

    def __getitem__(self, index: DatasetIndex):
        sample = self.get(index.recording_id, index.start, index.end)

        # apply transform
        if self.transform is not None:
            sample = self.transform(sample)

        return sample

    def __len__(self):
        raise NotImplementedError("Length of dataset is not defined")

    def __iter__(self):
        raise NotImplementedError("Iteration over dataset is not defined")

    def __repr__(self):
        return f"Dataset(root={self.root}, config={self.config}, split={self.split})"
